from clinicedc_constants import NO, YES
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.template.loader import render_to_string
from django.urls import NoReverseMatch, reverse
from django.utils.translation import gettext
from django_audit_fields import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...models import Allocation
from ...utils import may_delete_allocation
from ..list_filters import AssignmentListFilter
from ..model_admin_mixin import ModelAdminMixin
from ..remove_fields_for_blinded_users import remove_fields_for_blinded_users


class HasStockListFilter(SimpleListFilter):
    title = "Stock"
    parameter_name = "has_stock"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(stock__isnull=False)
            elif self.value() == NO:
                qs = queryset.filter(stock__isnull=True)
        return qs


class TransferredFilter(SimpleListFilter):
    title = "Transferred"
    parameter_name = "transferred"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(stock__stocktransferitem__isnull=False)
            elif self.value() == NO:
                qs = queryset.filter(stock__stocktransferitem__isnull=True)
        return qs


class DispensedFilter(SimpleListFilter):
    title = "Dispensed"
    parameter_name = "dispensed"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(stock__dispenseitem__isnull=False)
            elif self.value() == NO:
                qs = queryset.filter(stock__dispenseitem_isnull=True)
        return qs


class HasStockFilter(SimpleListFilter):
    title = "Orphaned"
    parameter_name = "orphaned"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(stock__isnull=True)
            elif self.value() == NO:
                qs = queryset.filter(stock__isnull=False)
        return qs


@admin.register(Allocation, site=edc_pharmacy_admin)
class AllocationAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Allocations"
    change_form_title = "Pharmacy: Allocation"
    change_list_note = "T=Transferred to location, CL=Confirmed at location, D=Dispensed"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    actions = ("delete_selected",)

    ordering = (
        "registered_subject__subject_identifier",
        "allocation_datetime",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "allocation_identifier",
                    "allocation_datetime",
                    "stock_request_item",
                    "registered_subject",
                    "code",
                    "allocated_by",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "allocation_date",
        "transferred",
        "confirmed_at_location",
        "dispensed",
        "dashboard",
        "stock_changelist",
        "stock_request_changelist",
        "stock_product",
        "stock_container",
        "assignment",
        "allocated_by",
    )

    list_filter = (
        "stock__location",
        AssignmentListFilter,
        ("allocation_datetime", DateRangeFilterBuilder()),
        HasStockListFilter,
        "stock__in_transit",
        "stock__confirmed_at_location",
        "stock__dispensed",
        "allocated_by",
        HasStockFilter,
    )

    search_fields = (
        "id",
        "code",
        "stock_request_item__id",
        "stock_request_item__stock_request__id",
        "stock_request_item__stock_request__request_identifier",
        "registered_subject__subject_identifier",
    )

    readonly_fields = (
        "assignment",
        "allocation_identifier",
        "allocation_datetime",
        "registered_subject",
        "stock_request_item",
        "code",
        "allocated_by",
    )

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        return remove_fields_for_blinded_users(request, fields)

    def get_list_filter(self, request):
        fields = super().get_list_filter(request)
        return remove_fields_for_blinded_users(request, fields)

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        return remove_fields_for_blinded_users(request, fields)

    @admin.display(description="ALLOCATION #", ordering="allocation_identifier")
    def identifier(self, obj):
        return obj.allocation_identifier.split("-")[0]

    @admin.display(description="Allocation date", ordering="allocation_datetime")
    def allocation_date(self, obj):
        return to_local(obj.allocation_datetime).date()

    @admin.display(description="T", boolean=True)
    def transferred(self, obj):
        return obj.stock.in_transit

    @admin.display(description="CL", boolean=True)
    def confirmed_at_location(self, obj):
        return obj.stock.confirmed_at_location

    @admin.display(description="D", boolean=True)
    def dispensed(self, obj):
        return obj.stock.dispensed

    @admin.display(description="Product", ordering="stock__product")
    def stock_product(self, obj):
        return obj.stock.product.name

    @admin.display(description="Assignment", ordering="stock__product__assignment")
    def assignment(self, obj):
        return obj.stock.product.assignment

    @admin.display(description="Product", ordering="stock__product")
    def stock_container(self, obj):
        return obj.stock.container

    @admin.display(
        description="Request #",
        ordering="stock_request_item__stock_request__request_identifier",
    )
    def stock_request_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stockrequest_changelist")
        url = f"{url}?q={obj.stock_request_item.stock_request.id}"
        context = dict(
            url=url,
            label=f"{obj.stock_request_item.stock_request.request_identifier}",
            title="Go to stock request",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock #")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=f"{obj.stock.code}", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Subject #", ordering="registered_subject__subject_identifier")
    def dashboard(self, obj=None, label=None):  # noqa: ARG002
        context = {}
        try:
            url = reverse(
                self.get_subject_dashboard_url_name(),
                kwargs={"subject_identifier": obj.registered_subject.subject_identifier},
            )
        except NoReverseMatch:
            url = None
        else:
            context = dict(
                title=gettext("Go to subject dashboard"),
                url=url,
                label=obj.registered_subject.subject_identifier,
            )
        if url:
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    def has_delete_permission(self, request, obj=None):
        """Hack to block delete if the stock instance has been
        transferred to a location other than CENTRAL.

        Block as if the user does not have permissions to
        delete. An additional message shows why.

        User requires explicit permission to delete this model
        (which is not the default for any role).
        """
        return (
            super().has_delete_permission(request, obj)
            if may_delete_allocation(self, request, obj)
            else False
        )



from textwrap import wrap

from django.apps import apps as django_apps
from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django.utils.html import format_html
from django_audit_fields import audit_fieldset_tuple

from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_sites.admin import SiteModelAdminMixin
from edc_visit_schedule.fieldsets import visit_schedule_only_fieldset_tuple

from ..admin_site import edc_form_runners_admin
from ..models import Issue
from .actions import issue_refresh


@admin.register(Issue, site=edc_form_runners_admin)
class IssueAdmin(
    SiteModelAdminMixin,
    ModelAdminSubjectDashboardMixin,
    admin.ModelAdmin,
):
    list_per_page = 15
    show_cancel = True
    actions = [issue_refresh]

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "verbose_name",
                    "subject_identifier",
                    "visit_code",
                    "visit_code_sequence",
                    "src_report_datetime",
                    "src_revision",
                    "src_id",
                    "site",
                )
            },
        ),
        (
            "Message",
            {
                "fields": (
                    "field_name",
                    "label_lower",
                    "panel_name",
                    "message",
                    "raw_message",
                )
            },
        ),
        (
            "Session",
            {"fields": ("session_id", "session_datetime")},
        ),
        visit_schedule_only_fieldset_tuple,
        audit_fieldset_tuple,
    )

    list_display = (
        "subject_identifier",
        "dashboard",
        "document",
        "error_msg",
        "field_name",
        "response",
        "visit",
        "src_report_datetime",
        "src_user_created",
        "src_user_modified",
        "src_created_datetime",
        "src_modified_datetime",
    )

    list_filter = (
        "visit_code",
        "visit_code_sequence",
        "panel_name",
        "verbose_name",
        "field_name",
        "short_message",
        "session_id",
        "session_datetime",
        "site",
        "src_user_created",
        "src_user_modified",
        "src_report_datetime",
        "src_created_datetime",
        "src_modified_datetime",
    )

    readonly_fields = (
        "field_name",
        "label_lower",
        "message",
        "panel_name",
        "raw_message",
        "schedule_name",
        "session_datetime",
        "session_id",
        "site",
        "src_created_datetime",
        "src_id",
        "src_modified_datetime",
        "src_report_datetime",
        "src_revision",
        "src_user_created",
        "src_user_modified",
        "src_user_modified",
        "subject_identifier",
        "verbose_name",
        "visit_code",
        "visit_code_sequence",
        "visit_schedule_name",
    )

    search_fields = (
        "message",
        "label_lower",
        "field_name",
        "subject_identifier",
        "src_id",
    )

    @admin.display(description="Message", ordering="short_message")
    def error_msg(self, obj):
        return format_html("<BR>".join(wrap(obj.short_message, 45)).replace(" ", "&nbsp"))

    @admin.display(description="Visit", ordering="visit_code")
    def visit(self, obj):
        return f"{obj.visit_code}.{obj.visit_code_sequence}"

    @admin.display(description="Document", ordering="verbose_name")
    def document(self, obj):
        if obj.panel_name:
            return f"{obj.verbose_name.title()} {obj.panel_name.replace('_', ' ').title()}"
        return obj.verbose_name

    def get_subject_dashboard_url_kwargs(self, obj) -> dict:
        appointment_model_cls = django_apps.get_model("edc_appointment.appointment")
        try:
            appointment = appointment_model_cls.objects.get(
                subject_identifier=obj.subject_identifier,
                visit_schedule_name=obj.visit_schedule_name,
                schedule_name=obj.schedule_name,
                visit_code=obj.visit_code,
                visit_code_sequence=obj.visit_code_sequence,
            )
        except ObjectDoesNotExist:
            opts = dict(subject_identifier=obj.subject_identifier)
        else:
            opts = dict(
                subject_identifier=appointment.subject_identifier,
                appointment=str(appointment.id),
            )
        return opts



from sequences import get_next_value

from .storage_bin import StorageBin


class StorageBinProxy(StorageBin):
    def _get_bin_identifier(self):
        return (
            f"{get_next_value(self._meta.proxy_for_model._meta.label_lower):06d}"
            if not self.id
            else None
        )

    class Meta:
        proxy = True
        verbose_name = "Central Storage bin"
        verbose_name_plural = "Central Storage bins"



from django.apps import apps as django_apps
from django.db import models
from django.utils import timezone

from edc_model.models import BaseUuidModel
from edc_sites.model_mixins import SiteModelMixin

from ..choices import NOTE_STATUSES


class NoteModelMixin(SiteModelMixin, BaseUuidModel):
    """Model mixin to link form (e.g. note) to a data query report,
    such as, unmanaged views.

    See also, NoteModelAdminMixin
    """

    report_model = models.CharField(max_length=150)

    report_datetime = models.DateTimeField(default=timezone.now)

    note = models.TextField(
        default="",
        blank=True,
    )

    status = models.CharField(max_length=25, choices=NOTE_STATUSES, default="", blank=False)

    @property
    def report_model_cls(self):
        return django_apps.get_model(self.report_model)

    class Meta:
        abstract = True



from .crf_case import CrfCase, CrfCaseError
from .crf_subquery import CrfSubquery
from .requisition_case import RequisitionCase
from .requisition_subquery import RequisitionSubquery
from .sql_view_generator import SqlViewGenerator



from .constants import CSV, STATA_10, STATA_13, STATA_14, STATA_15

EXPORT_FORMATS = (
    (CSV, "CSV (delimited by pipe `|`)"),
    (STATA_10, "Stata v10 or later"),
    (STATA_13, "Stata v13 or later"),
    (STATA_14, "Stata v14 or later"),
    (STATA_15, "Stata v15 or later"),
)



from django.db import models


class EdcPermissionsManager(models.Manager):
    def get_by_natural_key(self, pk):
        return self.get(id=pk)


class EdcPermissionsModelMixin(models.Model):
    objects = EdcPermissionsManager()

    def __str__(self):
        return self._meta.app_label

    def natural_key(self):
        return (self.id,)

    class Meta:
        abstract = True
        verbose_name = "Edc Permissions"
        verbose_name_plural = "Edc Permissions"



# Generated by Django 3.0.6 on 2020-07-29 20:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0016_auto_20200327_1435"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="datadictionary",
            options={
                "default_permissions": ("view", "export"),
                "ordering": ("model", "number", "prompt"),
                "verbose_name": "Data Dictionary Item",
                "verbose_name_plural": "Data Dictionary Items",
            },
        ),
        migrations.AlterModelOptions(
            name="datamanageruser",
            options={
                "default_permissions": ("view", "export"),
                "verbose_name": "Data Manager User",
                "verbose_name_plural": "Data Manager Users",
            },
        ),
        migrations.AlterModelOptions(
            name="dataquery",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Data Query",
                "verbose_name_plural": "Data Queries",
            },
        ),
        migrations.AlterModelOptions(
            name="queryrule",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("title",),
                "verbose_name": "Query Rule",
                "verbose_name_plural": "Query Rules",
            },
        ),
        migrations.AlterModelOptions(
            name="querysubject",
            options={
                "default_permissions": ("view", "export"),
                "verbose_name": "Subject",
                "verbose_name_plural": "Subjects",
            },
        ),
        migrations.AlterModelOptions(
            name="queryuser",
            options={
                "default_permissions": ("view", "export"),
                "verbose_name": "Data Query User",
                "verbose_name_plural": "Data Query Users",
            },
        ),
        migrations.AlterModelOptions(
            name="queryvisitschedule",
            options={
                "default_permissions": ("view", "export"),
                "verbose_name": "Visit Schedule",
                "verbose_name_plural": "Visit Schedule",
            },
        ),
        migrations.AlterModelOptions(
            name="requisitionpanel",
            options={
                "default_permissions": ("view", "export"),
                "verbose_name": "Requisition Panel",
                "verbose_name_plural": "Requisition Panels",
            },
        ),
    ]



from django.urls import path

from .admin_site import edc_pharmacy_admin
from .constants import CENTRAL_LOCATION
from .views import (
    AddToStorageBinView,
    AllocateToSubjectView,
    CeleryTaskStatusView,
    ConfirmaAtLocationView,
    ConfirmStockFromQuerySetView,
    DispenseView,
    HomeView,
    MoveToStorageBinView,
    PrepareAndReviewStockRequestView,
    PrintLabelsView,
    TransferStockView,
    get_stock_transfers_view,
    print_stock_transfer_manifest_view,
    print_stock_view,
)

app_name = "edc_pharmacy"


urlpatterns = [
    path(
        "dispense/<int:location_id>/<uuid:formulation_id>/"
        "<str:subject_identifier>/<int:container_count>/",
        DispenseView.as_view(),
        name="dispense_url",
    ),
    path("get-stock-transfers/", get_stock_transfers_view, name="get_stock_transfers_url"),
    path(
        "confirm-at-location/<uuid:session_uuid>/<str:stock_transfer_identifier>/"
        "<int:location_id>/<int:items_to_scan>/",
        ConfirmaAtLocationView.as_view(),
        name="confirm_at_location_url",
    ),
    path(
        "confirm-at-location/<str:stock_transfer_identifier>/"
        "<int:location_id>/<int:items_to_scan>/",
        ConfirmaAtLocationView.as_view(),
        name="confirm_at_location_url",
    ),
    path(
        "confirm-at-location/<int:location_id>/<int:items_to_scan>/",
        ConfirmaAtLocationView.as_view(),
        name="confirm_at_location_url",
    ),
    path(
        "confirm-at-location/<int:site_id>/",
        ConfirmaAtLocationView.as_view(),
        name="confirm_at_location_url",
    ),
    path(
        "confirm-at-location/<str:location_name>/",
        ConfirmaAtLocationView.as_view(),
        name="confirm_at_location_url",
    ),
    path(
        "review-stock-request/<uuid:stock_request>/<uuid:session_uuid>/",
        PrepareAndReviewStockRequestView.as_view(),
        name="review_stock_request_url",
    ),
    path(
        "allocate/<uuid:stock_request>/<uuid:assignment>/",
        AllocateToSubjectView.as_view(),
        name="allocate_url",
    ),
    path(
        "allocate/<uuid:stock_request>/<uuid:assignment>",
        AllocateToSubjectView.as_view(),
        name="allocate_url",
    ),
    path(
        "confirm-stock-qs/<uuid:session_uuid>/",
        ConfirmStockFromQuerySetView.as_view(),
        name="confirm_stock_from_queryset_url",
    ),
    path(
        "allocate/<uuid:stock_request>/",
        AllocateToSubjectView.as_view(),
        name="allocate_url",
    ),
    path(
        "transfer-stock/<uuid:stock_transfer>/",
        TransferStockView.as_view(),
        name="transfer_stock_url",
    ),
    path(
        "review-stock-request/<uuid:stock_request>/",
        PrepareAndReviewStockRequestView.as_view(),
        name="review_stock_request_url",
    ),
    path(
        "print-labels/<str:model>/<uuid:session_uuid>/<str:label_configuration>",
        PrintLabelsView.as_view(),
        name="print_labels_url",
    ),
    path(
        "print-labels/<str:model>/<uuid:session_uuid>/",
        PrintLabelsView.as_view(),
        name="print_labels_url",
    ),
    path(
        "manifest/<uuid:stock_transfer>/",
        print_stock_transfer_manifest_view,
        name="generate_manifest",
    ),
    path("stock_report/<uuid:session_uuid>/", print_stock_view, name="stock_report"),
    path(
        "add-to-storage-bin/<uuid:storage_bin>/<int:items_to_scan>/",
        AddToStorageBinView.as_view(),
        name="add_to_storage_bin_url",
    ),
    path(
        "add-to-storage-bin/<uuid:storage_bin>/",
        AddToStorageBinView.as_view(),
        name="add_to_storage_bin_url",
    ),
    path(
        "move-to-storage-bin/<uuid:storage_bin>/<int:items_to_scan>/",
        MoveToStorageBinView.as_view(),
        name="move_to_storage_bin_url",
    ),
    path(
        "move-to-storage-bin/<uuid:storage_bin>/",
        MoveToStorageBinView.as_view(),
        name="move_to_storage_bin_url",
    ),
    path(
        "task-status/<uuid:task_id>/",
        CeleryTaskStatusView.as_view(),
        name="celery_task_status_url",
    ),
    path(
        "stock-transfer-confirmation/",
        ConfirmaAtLocationView.as_view(),
        name="confirm_at_location_url",
    ),
    path(
        "dispense/",
        DispenseView.as_view(),
        name="dispense_url",
    ),
    path("admin/", edc_pharmacy_admin.urls),
    # path("admin/history/", edc_pharmacy_history_admin.urls),
    path("", HomeView.as_view(), name="home_url"),
]
1



from clinicedc_constants import NEW, NO, NOT_APPLICABLE, NULL_STRING, PARTIAL, RECEIVED, YES
from django.contrib.admin import SimpleListFilter
from django.contrib.sites.shortcuts import get_current_site
from django.db.models import Count, F, Q
from django.utils.translation import gettext as _
from edc_constants.choices import YES_NO, YES_NO_NA

from ..models import Medication, Rx
from ..utils import blinded_user


class MedicationsListFilter(SimpleListFilter):
    title = "Medication"
    parameter_name = "medication_name"

    def lookups(self, request, model_admin) -> tuple[tuple[str, str], ...]:  # noqa: ARG002
        medications = [
            (medication.name, medication.name.replace("_", " ").title())
            for medication in Medication.objects.all().order_by("name")
        ]
        medications.append(("none", "None"))
        return tuple(medications)

    def queryset(self, request, queryset):  # noqa: ARG002
        """Returns a queryset if the Medication name is in the list of sites"""
        qs = None
        if self.value():
            if self.value() == "none":
                qs = Rx.objects.filter(
                    medications__isnull=True, site=get_current_site(request)
                )
            else:
                qs = Rx.objects.filter(
                    medications__name__in=[self.value()], site=get_current_site(request)
                )
        return qs


class ConfirmedListFilter(SimpleListFilter):
    title = "Confirmed"
    parameter_name = "confirmation"

    def lookups(self, request, model_admin) -> tuple[tuple[str, str], ...]:  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(confirmed=True)
            elif self.value() == NO:
                qs = queryset.filter(confirmed=False)
        return qs


class AllocationListFilter(SimpleListFilter):
    title = "Allocated"
    parameter_name = "allocated"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO_NA

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(
                    from_stock__isnull=False,
                    allocation__isnull=False,
                    container__may_request_as=True,
                )
            elif self.value() == NO:
                qs = queryset.filter(
                    from_stock__isnull=False,
                    allocation__isnull=True,
                    container__may_request_as=True,
                )
            elif self.value() == NOT_APPLICABLE:
                qs = queryset.filter(
                    from_stock__isnull=True,
                    allocation__isnull=True,
                    container__may_request_as=False,
                )
        return qs


class StockItemAllocationListFilter(SimpleListFilter):
    title = "Allocated"
    parameter_name = "allocated"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(allocation__isnull=False)
            elif self.value() == NO:
                qs = queryset.filter(allocation__isnull=True)
        return qs


class StockItemTransferredListFilter(SimpleListFilter):
    title = "Transferred"
    parameter_name = "transferred"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        """Note that the only way a stock item can be tranferred is if it
        first is allocated to a subject.
        """
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(
                    allocation__stock__in_transit=True,
                )
            elif self.value() == NO:
                qs = queryset.filter(
                    Q(allocation__isnull=True) | Q(allocation__stock__in_transit=False),
                )
        return qs


class StockItemConfirmedAtLocationListFilter(SimpleListFilter):
    title = "Confirmed at location"
    parameter_name = "confirmed_at_location"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        """Note that the only way a stock item can be confirmed_at_location is if it
        first is allocated to a subject.
        """
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(
                    allocation__stock__confirmed_at_location=True,
                )
            elif self.value() == NO:
                qs = queryset.filter(
                    Q(allocation__isnull=True)
                    | Q(allocation__stock__confirmed_at_location=False),
                )
        return qs


class TransferredListFilter(SimpleListFilter):
    title = "Transferred"
    parameter_name = "transferred"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO_NA

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(
                    container__may_request_as=True,
                    allocation__stock_request_item__stock_request__location=F("location"),
                )
            elif self.value() == NO:
                qs = queryset.filter(
                    ~Q(allocation__stock_request_item__stock_request__location=F("location")),
                    container__may_request_as=True,
                )
            elif self.value() == NOT_APPLICABLE:
                qs = queryset.filter(allocation__isnull=True, container__may_request_as=False)
        return qs


class AssignmentListFilter(SimpleListFilter):
    title = "Assignment"
    parameter_name = "assignment"
    lookup_str = "assignment__name"

    def lookups(self, request, model_admin):
        groupby = model_admin.model.objects.values(self.lookup_str).annotate(
            count=Count(self.lookup_str)
        )
        if not blinded_user(request):
            choices = [
                (name, name or "None")
                for name in [ann.get(self.lookup_str) for ann in groupby]
            ]
            return tuple(choices)
        return ("****", "****"), ("****", "****")

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            qs = queryset.filter(**{self.lookup_str: self.value()})
        return qs


class ProductAssignmentListFilter(AssignmentListFilter):
    title = "Assignment"
    parameter_name = "product_assignment"
    lookup_str = "product__assignment__name"


class HasOrderNumFilter(SimpleListFilter):
    title = "Has Order #"
    parameter_name = "has_order_num"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            isnull = self.value() == NO
            qs = queryset.filter(receive_item__order_item__order__isnull=isnull)
        return qs


class HasReceiveNumFilter(SimpleListFilter):
    title = "Has Receive #"
    parameter_name = "has_receive_num"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            isnull = self.value() == NO
            qs = queryset.filter(receive_item__receive__isnull=isnull)
        return qs


class HasRepackNumFilter(SimpleListFilter):
    title = "Has Repack #"
    parameter_name = "has_repack_num"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            isnull = self.value() == NO
            qs = queryset.filter(repack_request__isnull=isnull)
        return qs


class TransferredFilter(SimpleListFilter):
    title = "Transferred to site"
    parameter_name = "transferred"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            opts = dict(
                from_stock__isnull=False,
                confirmation__isnull=False,
                allocation__isnull=False,
            )
            if self.value() == YES:
                qs = queryset.filter(stocktransferitem__isnull=False, **opts)
            elif self.value() == NO:
                qs = queryset.filter(stocktransferitem__isnull=True, **opts)
        return qs


class ConfirmedAtLocationFilter(SimpleListFilter):
    title = "Confirmed at location"
    parameter_name = "confirmed_at_location"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            opts = dict(
                from_stock__isnull=False,
                confirmation__isnull=False,
                allocation__isnull=False,
                stocktransferitem__isnull=False,
            )
            if self.value() == YES:
                qs = queryset.filter(confirmationatlocationitem__isnull=False, **opts)
            elif self.value() == NO:
                qs = queryset.filter(confirmationatlocationitem__isnull=True, **opts)
        return qs


class StoredAtSiteFilter(SimpleListFilter):
    title = "Stored at location"
    parameter_name = "stored_at_location_now"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO_NA

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(
                    from_stock__isnull=False,
                    confirmation__isnull=False,
                    allocation__isnull=False,
                    stocktransferitem__isnull=False,
                    confirmationatlocationitem__isnull=False,
                    storagebinitem__isnull=False,
                    dispenseitem__isnull=True,
                )
            elif self.value() == NO:
                qs = queryset.filter(
                    from_stock__isnull=False,
                    confirmation__isnull=False,
                    allocation__isnull=False,
                    stocktransferitem__isnull=False,
                    confirmationatlocationitem__isnull=False,
                    storagebinitem__isnull=True,
                    dispenseitem__isnull=True,
                )
            elif self.value() == NOT_APPLICABLE:
                qs = queryset.filter(
                    from_stock__isnull=False,
                    confirmation__isnull=False,
                    allocation__isnull=False,
                    stocktransferitem__isnull=False,
                    confirmationatlocationitem__isnull=False,
                    storagebinitem__isnull=True,
                    dispenseitem__isnull=False,
                )
        return qs


class DispensedFilter(SimpleListFilter):
    title = "Dispensed"
    parameter_name = "dispensed"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            opts = dict(
                from_stock__isnull=False,
                confirmation__isnull=False,
                allocation__isnull=False,
                stocktransferitem__isnull=False,
                confirmationatlocationitem__isnull=False,
            )
            if self.value() == YES:
                qs = queryset.filter(dispenseitem__isnull=False, **opts)
            elif self.value() == NO:
                qs = queryset.filter(dispenseitem__isnull=True, **opts)
        return qs


class ReturnedFilter(SimpleListFilter):
    title = "Returned"
    parameter_name = "returned"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            opts = dict(
                from_stock__isnull=False,
                confirmation__isnull=False,
                allocation__isnull=False,
                stocktransferitem__isnull=False,
                confirmationatlocationitem__isnull=False,
                dispenseitem__isnull=True,
            )
            if self.value() == YES:
                qs = queryset.filter(stockreturnitem__isnull=False, **opts)
            elif self.value() == NO:
                qs = queryset.filter(stockreturnitem__isnull=True, **opts)
        return qs


class HasCodesListFilter(SimpleListFilter):
    title = "Has codes"
    parameter_name = "has_codes"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.exclude(codes=NULL_STRING)
            elif self.value() == NO:
                qs = queryset.filter(codes=NULL_STRING)
        return qs


class StockRequestItemPendingListFilter(SimpleListFilter):
    title = "Request Status"
    parameter_name = "item_status"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (
            ("not_allocated", "Not allocated"),
            ("allocated_only", "Allocated only"),
            ("transferred", "Allocation and transferred"),
        )

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == "not_allocated":
                qs = queryset.filter(allocation__isnull=True)
            elif self.value() == "allocated_only":
                qs = queryset.filter(
                    Q(allocation__isnull=False)
                    & Q(allocation__stock__stocktransferitem__isnull=True)
                    & ~Q(allocation__stock__location=F("stock_request__location"))
                )
            elif self.value() == "transferred":
                qs = queryset.filter(
                    Q(allocation__isnull=False)
                    & Q(allocation__stock__stocktransferitem__isnull=False)
                    & Q(allocation__stock__location=F("stock_request__location"))
                )
        return qs


class OrderItemStatusListFilter(SimpleListFilter):
    title = "Status"
    parameter_name = "order_item_status"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (
            (NEW, _("New")),
            (PARTIAL, _("Partial")),
            (RECEIVED, _("Received")),
        )

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == RECEIVED:
                qs = queryset.filter(unit_qty=0)
            elif self.value() == PARTIAL:
                qs = queryset.filter(Q(unit_qty_ordered__gt=F("unit_qty")) & ~Q(unit_qty=0))
            elif self.value() == NEW:
                qs = queryset.filter(unit_qty=F("unit_qty_ordered"))
        return qs


class DecantedListFilter(SimpleListFilter):
    title = "Decanted"
    parameter_name = "decanted"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return YES_NO

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(from_stock__isnull=False)
            elif self.value() == NO:
                qs = queryset.filter(from_stock__isnull=True)
        return qs


class StorageBinItemDispensedFilter(SimpleListFilter):
    title = "Dispensed"
    parameter_name = "dispensed"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(
                    stock__dispenseitem__isnull=False,
                )
            elif self.value() == NO:
                qs = queryset.filter(
                    stock__dispenseitem__isnull=True,
                )
        return qs



from edc_lab import RequisitionPanel

from .constants import (
    BLOOD_GLUCOSE,
    BLOOD_GLUCOSE_POC,
    CD4,
    CHEMISTRY_LFT,
    CHEMISTRY_LIPIDS,
    CHEMISTRY_RFT,
    CHOL,
    FBC,
    HBA1C,
    HBA1C_POC,
    HDL,
    INSULIN,
    LDL,
    LFT,
    LIPIDS,
    RFT,
    SPUTUM,
    TRIG,
    VL,
)
from .processing_profiles import (
    blood_glucose_processing,
    cd4_processing,
    fbc_processing,
    hba1c_processing,
    insulin_processing,
    lft_processing,
    lipids_processing,
    poc_processing,
    rft_processing,
    sputum_processing,
    vl_processing,
)

hba1c_panel = RequisitionPanel(
    name=HBA1C,
    verbose_name="HbA1c (Venous)",
    processing_profile=hba1c_processing,
    abbreviation="HBA1C",
    utest_ids=(("hba1c", "HbA1c"),),
)

hba1c_poc_panel = RequisitionPanel(
    name=HBA1C_POC,
    verbose_name="HbA1c (POC)",
    abbreviation="HBA1C_POC",
    processing_profile=poc_processing,
    utest_ids=(("hba1c", "HbA1c"),),
)


fbc_panel = RequisitionPanel(
    name=FBC,
    verbose_name="Full Blood Count",
    processing_profile=fbc_processing,
    abbreviation="FBC",
    utest_ids=(
        ("haemoglobin", "Haemoglobin"),
        "hct",
        "rbc",
        "wbc",
        "platelets",
        "mcv",
        "mch",
        "mchc",
    ),
)

blood_glucose_panel = RequisitionPanel(
    name=BLOOD_GLUCOSE,
    verbose_name="Blood Glucose (Venous)",
    abbreviation="BGL",
    processing_profile=blood_glucose_processing,
    utest_ids=(("glucose", "Glucose"),),
)

blood_glucose_poc_panel = RequisitionPanel(
    name=BLOOD_GLUCOSE_POC,
    verbose_name="Blood Glucose (POC)",
    abbreviation="BGL-POC",
    processing_profile=poc_processing,
    utest_ids=(("glucose", "Glucose"),),
)

cd4_panel = RequisitionPanel(
    name=CD4,
    verbose_name="CD4",
    abbreviation="CD4",
    processing_profile=cd4_processing,
    utest_ids=("cd4",),
)
vl_panel = RequisitionPanel(
    name=VL,
    verbose_name="Viral Load",
    abbreviation="VL",
    processing_profile=vl_processing,
    utest_ids=("vl",),
)


rft_panel = RequisitionPanel(
    name=CHEMISTRY_RFT,
    verbose_name="Chemistry: Renal Function Tests",
    abbreviation=RFT,
    processing_profile=rft_processing,
    utest_ids=("urea", "creatinine", "uric_acid", "egfr", "egfr_drop"),
)

lipids_panel = RequisitionPanel(
    name=CHEMISTRY_LIPIDS,
    verbose_name="Chemistry: Lipids",
    abbreviation=LIPIDS,
    processing_profile=lipids_processing,
    utest_ids=(LDL, HDL, TRIG, CHOL),
)

lft_panel = RequisitionPanel(
    name=CHEMISTRY_LFT,
    verbose_name="Chemistry: Liver Function Tests",
    abbreviation=LFT,
    processing_profile=lft_processing,
    utest_ids=("ast", "alt", "alp", "amylase", "ggt", "albumin"),
)

insulin_panel = RequisitionPanel(
    name=INSULIN,
    verbose_name="Insulin",
    abbreviation="INS",
    processing_profile=insulin_processing,
    utest_ids=("ins",),
)

sputum_panel = RequisitionPanel(
    name=SPUTUM,
    verbose_name="Sputum",
    abbreviation="SPM",
    processing_profile=sputum_processing,
    utest_ids=(),
)



from __future__ import annotations

from django.apps import apps as django_apps

from .exceptions import (
    FormRunnerError,
    FormRunnerImproperlyConfigured,
    FormRunnerModelAdminNotFound,
    FormRunnerModelFormNotFound,
)
from .get_form_runner import get_form_runner

__all__ = ["run_form_runners"]

from .models import Issue


def run_form_runners(
    app_labels: list[str] | None = None, model_names: list[str] | None = None
):
    model_names = model_names or []
    if app_labels:
        for app_config in django_apps.get_app_configs():
            if app_config.name in app_labels:
                for model_cls in app_config.get_models():
                    if not model_cls._meta.label_lower.split(".")[1].startswith("historical"):
                        model_names.append(model_cls._meta.label_lower)
    if not model_names:
        raise FormRunnerError("Nothing to do.")
    for model_name in model_names:
        print(model_name)
        Issue.objects.filter(label_lower=model_name).delete()
        try:
            get_form_runner(model_name, verbose=True).run_all()
        except (
            FormRunnerImproperlyConfigured,
            FormRunnerModelAdminNotFound,
            FormRunnerModelFormNotFound,
        ) as e:
            print(f"{e} See {model_name}.")
        except AttributeError as e:
            print(f"{e} See {model_name}.")



from django.conf import settings
from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin


class HomeView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = "edc_data_manager/home.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "data_manager_home"



from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin

from ..managers import AppointmentManager
from ..model_mixins import AppointmentModelMixin


class Appointment(AppointmentModelMixin, SiteModelMixin, BaseUuidModel):
    on_site = CurrentSiteManager()

    objects = AppointmentManager()

    history = HistoricalRecords()

    def natural_key(self) -> tuple:
        return (
            self.subject_identifier,
            self.visit_schedule_name,
            self.schedule_name,
            self.visit_code,
            self.visit_code_sequence,
        )

    # noinspection PyTypeHints
    natural_key.dependencies = ("sites.Site",)

    class Meta(AppointmentModelMixin.Meta, SiteModelMixin.Meta, BaseUuidModel.Meta):
        indexes = (*AppointmentModelMixin.Meta.indexes, *BaseUuidModel.Meta.indexes)



UNBLINDING_REQUESTORS_ROLE = "UNBLINDING_REQUESTORS_ROLE"
UNBLINDING_REVIEWERS_ROLE = "UNBLINDING_REVIEWERS_ROLE"

UNBLINDING_REQUESTORS = "UNBLINDING_REQUESTORS"
UNBLINDING_REVIEWERS = "UNBLINDING_REVIEWERS"

unblinding_requestors = [
    "edc_unblinding.add_unblindingrequest",
    "edc_unblinding.change_unblindingrequest",
    "edc_unblinding.delete_unblindingrequest",
    "edc_unblinding.view_unblindingrequest",
    "edc_unblinding.view_historicalunblindingrequest",
    "edc_unblinding.view_unblindingrequestoruser",
]


unblinding_reviewers = [
    "edc_unblinding.add_unblindingreview",
    "edc_unblinding.change_unblindingreview",
    "edc_unblinding.delete_unblindingreview",
    "edc_unblinding.view_unblindingreview",
    "edc_unblinding.view_historicalunblindingreview",
    "edc_unblinding.view_unblindingrevieweruser",
]



from edc_model_admin.dashboard import ModelAdminCrfDashboardMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_model_admin.mixins import ModelAdminProtectPiiMixin
from edc_sites.admin import SiteModelAdminMixin

from .crf_status_model_admin_mixin import CrfStatusModelAdminMixin


class CrfModelAdmin(
    ModelAdminProtectPiiMixin,  # must remain first
    SiteModelAdminMixin,
    CrfStatusModelAdminMixin,
    ModelAdminCrfDashboardMixin,
    SimpleHistoryAdmin,
):
    pass



# Generated by Django 3.2.13 on 2022-09-24 21:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0031_auto_20220922_2236"),
        ("edc_offstudy", "0014_auto_20220918_0508"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalsubjectoffstudy",
            name="action_identifier",
            field=models.CharField(blank=True, db_index=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="historicalsubjectoffstudy",
            name="action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectoffstudy",
            name="action_item_reason",
            field=models.TextField(editable=False, null=True),
        ),
        migrations.AddField(
            model_name="historicalsubjectoffstudy",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectoffstudy",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectoffstudy",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectoffstudy",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AddField(
            model_name="subjectoffstudy",
            name="action_identifier",
            field=models.CharField(blank=True, max_length=50, null=True, unique=True),
        ),
        migrations.AddField(
            model_name="subjectoffstudy",
            name="action_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AddField(
            model_name="subjectoffstudy",
            name="action_item_reason",
            field=models.TextField(editable=False, null=True),
        ),
        migrations.AddField(
            model_name="subjectoffstudy",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="subjectoffstudy",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AddField(
            model_name="subjectoffstudy",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="subjectoffstudy",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.actionitem",
            ),
        ),
    ]



from django.conf import settings

from .dashboard_templates import dashboard_templates
from .middleware_mixins import EdcTemplateMiddlewareMixin
from .url_names import url_names


class DashboardMiddleware(EdcTemplateMiddlewareMixin):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.check_for_required_request_attrs(request)
        return self.get_response(request)

    def process_view(self, request, *args):
        """Adds/Updates references to urls and templates."""
        request.url_name_data.update(**url_names.registry)
        template_data = getattr(settings, "DASHBOARD_BASE_TEMPLATES", {})
        template_data.update(**dashboard_templates)
        request.template_data.update(**template_data)

    def process_template_response(self, request, response):
        if getattr(response, "context_data", None):
            response.context_data.update(**request.url_name_data)
            response.context_data.update(**request.template_data)
        return response



from dataclasses import dataclass, field

import sqlglot
from django.apps import apps as django_apps
from django.db import OperationalError, connection

from .crf_subquery import CrfSubquery


class CrfCaseError(Exception):
    pass


@dataclass(kw_only=True)
class CrfCase:
    label: str = None
    dbtable: str = None
    label_lower: str = None
    fld_name: str | None = None
    where: str | None = None
    list_tables: list[tuple[str, str, str]] | None = field(default_factory=list)
    subjectvisit_dbtable: str | None = None

    @property
    def sql(self):
        sql = CrfSubquery(**self.__dict__).sql
        vendor = "postgres" if connection.vendor.startswith("postgres") else connection.vendor
        return sqlglot.transpile(sql, read="mysql", write=vendor)[0]

    @property
    def model_cls(self):
        return django_apps.get_model(self.label_lower)

    def fetchall(self):
        with connection.cursor() as cursor:
            try:
                cursor.execute(self.sql)
            except OperationalError as e:
                raise CrfCaseError(f"{e}. See {self}.") from e
            return cursor.fetchall()



from django.apps import apps as django_apps
from django.core.exceptions import ValidationError


class DevicePermissionAddError(ValidationError):
    pass


class DevicePermissionChangeError(ValidationError):
    pass


class DevicePermissionError(Exception):
    pass


class BaseDevicePermission:
    label = None
    exception_cls = None
    device_roles = []
    device_ids = []

    def __init__(self, model=None, device_roles=None, device_ids=None):
        self.model = model
        self.device_roles = device_roles or self.device_roles
        self.device_ids = device_ids or self.device_ids

    def __repr__(self):
        return f"{self.__class__.__name__}({self.model},{self.device_roles})"

    def __str__(self):
        roles = ",".join(self.device_roles)
        return f"{self.model} {self.label} {roles}"

    def model_operation(self, model_obj=None, **kwargs):
        """Override."""
        raise DevicePermissionError("Method not implemented")

    @property
    def _permit_model_operation(self):
        app_config = django_apps.get_app_config("edc_device")
        if (
            app_config.device_role in self.device_roles
            or app_config.device_id in self.device_ids
        ):
            return True
        return False

    def check(self, model_obj=None, err_message=None, **kwargs):
        if not self.model:
            self.model = model_obj._meta.label_lower
        if model_obj._meta.label_lower == self.model:
            if (
                self.model_operation(model_obj=model_obj, **kwargs)
                and not self._permit_model_operation
            ):
                app_config = django_apps.get_app_config("edc_device")
                err_message = err_message or (
                    f"Device role may not {self.label.lower()} "
                    f"'{model_obj._meta.verbose_name}'."
                )
                raise self.exception_cls(
                    f"Device/Role has insufficient permissions for action. "
                    f"Got {err_message}. Device role is {app_config.device_role}.",
                    code=f"{self.label}_permission",
                )


class DeviceAddPermission(BaseDevicePermission):
    label = "ADD"
    exception_cls = DevicePermissionAddError

    def model_operation(self, model_obj=None, **kwargs):
        """Returns ADD if this is an add model."""
        if not model_obj.id:
            return self.label
        return None


class DeviceChangePermission(BaseDevicePermission):
    label = "CHANGE"
    exception_cls = DevicePermissionChangeError

    def model_operation(self, model_obj=None, **kwargs):
        """Returns CHANGE if this is a change model."""
        if model_obj.id:
            return self.label
        return None


class DevicePermissions:
    """Container class for registered device permission instances."""

    def __init__(self):  # , *device_permissions):
        self._registry = []
        self.n = 0
        self.models = []

    #         for device_permission in device_permissions:
    #             self.register(device_permission)

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        try:
            item = self._registry[self.n]
            self.n += 1
        except IndexError:
            raise StopIteration
        return item

    def reset(self):
        self._registry = []
        self.n = 0
        self.models = []

    def register(self, device_permission):
        self._registry.append(device_permission)
        self._registry = list(set(self._registry))
        if device_permission.model:
            self.models.append(device_permission.model)
            self.models = list(set(self.models))

    def check(self, model_obj=None, **kwargs):
        for device_permission in self._registry:
            device_permission.check(model_obj=model_obj, **kwargs)


device_permissions = DevicePermissions()



# Generated by Django 6.0 on 2026-01-19 20:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0104_allocation_subject_identifier_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="allocation",
            name="code",
            field=models.CharField(
                blank=True,
                editable=False,
                help_text="A unique alphanumeric code",
                max_length=15,
                null=True,
                unique=True,
                verbose_name="Stock code",
            ),
        ),
        migrations.AddField(
            model_name="historicalallocation",
            name="code",
            field=models.CharField(
                blank=True,
                db_index=True,
                editable=False,
                help_text="A unique alphanumeric code",
                max_length=15,
                null=True,
                verbose_name="Stock code",
            ),
        ),
    ]



from .allocate_stock_to_subject import allocate_stock_to_subject
from .confirm_stock import (
    confirm_received_stock_action,
    confirm_repacked_stock_action,
    confirm_stock_from_queryset,
)
from .delete_items_for_stock_request import delete_items_for_stock_request_action
from .delete_order_items import delete_order_items_action
from .delete_receive_items import delete_receive_items_action
from .go_to_add_repack_request import go_to_add_repack_request_action
from .go_to_allocations import go_to_allocations
from .go_to_stock import go_to_stock
from .prepare_stock_request_items import prepare_stock_request_items_action
from .print_labels import (
    print_labels,
    print_labels_from_receive,
    print_labels_from_receive_item,
    print_labels_from_repack_request,
)
from .print_stock_labels import print_stock_labels
from .print_stock_report import print_stock_report_action
from .print_transfer_stock_manifest import print_transfer_stock_manifest_action
from .process_repack_request import process_repack_request_action
from .storage_bin import add_to_storage_bin_action
from .transfer_stock import transfer_stock_action



from __future__ import annotations

from pathlib import Path
from textwrap import wrap
from typing import TYPE_CHECKING

from clinicedc_constants import AE_WITHDRAWN, CLOSED, OPEN, OTHER, YES
from django import template
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.messages import ERROR
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import select_template
from django.utils import timezone
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from edc_action_item.utils import get_reference_obj
from edc_auth.constants import TMG_ROLE
from edc_model_admin.utils import add_to_messages_once
from edc_utils import escape_braces

from ..constants import (
    AE_TMG_ACTION,
    DEATH_REPORT_TMG_ACTION,
    DEATH_REPORT_TMG_SECOND_ACTION,
)
from ..utils import get_adverse_event_app_label, get_ae_model, has_valid_tmg_perms
from ..view_utils import TmgButton

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from edc_action_item.models import ActionItem
    from edc_model.models import BaseUuidModel

    from ..model_mixins import (
        AeFollowupModelMixin,
        AeInitialModelMixin,
        DeathReportModelMixin,
        DeathReportTmgModelMixin,
    )

    class DeathReportTmgModel(DeathReportTmgModelMixin, BaseUuidModel): ...

    class DeathReportTmgSecondModel(DeathReportTmgModelMixin, BaseUuidModel): ...

    class AeInitialModel(AeInitialModelMixin, BaseUuidModel): ...

    class AeFollowupModel(AeFollowupModelMixin, BaseUuidModel): ...

    class DeathReportModel(DeathReportModelMixin, BaseUuidModel): ...


register = template.Library()


def wrapx(text: str, length: int) -> str:
    if length:
        return "<BR>".join(wrap(text, length))
    return text


def select_ae_template(relative_path):
    """Returns a template object."""
    local_path = Path(get_adverse_event_app_label())
    default_path = Path("edc_adverse_event")
    return select_template(
        [
            str(local_path / relative_path),
            str(default_path / relative_path),
        ]
    )


def select_description_template(model) -> Path | str:
    """Returns a template name."""
    return select_ae_template(f"{model}_description.html").template.name


@register.inclusion_tag(select_description_template("aeinitial"), takes_context=True)
def format_ae_description(context, ae_initial, wrap_length):
    context["utc_date"] = timezone.now().date()
    context["SHORT_DATE_FORMAT"] = settings.SHORT_DATE_FORMAT
    context["OTHER"] = OTHER
    context["YES"] = YES
    context["ae_initial"] = ae_initial
    try:
        context["sae_reason"] = ae_initial.sae_reason.display_name
    except AttributeError:
        context["sae_reason"] = ""
    context["ae_description"] = ae_initial.ae_description
    return context


@register.inclusion_tag(select_description_template("aefollowup"), takes_context=True)
def format_ae_followup_description(context, ae_followup, wrap_length):
    context["AE_WITHDRAWN"] = AE_WITHDRAWN
    context["utc_date"] = timezone.now().date()
    context["SHORT_DATE_FORMAT"] = settings.SHORT_DATE_FORMAT
    context["OTHER"] = OTHER
    context["YES"] = YES
    context["ae_followup"] = ae_followup
    context["ae_initial"] = ae_followup.ae_initial
    try:
        context["sae_reason"] = mark_safe(  # noqa: S308
            wrapx(escape_braces(ae_followup.ae_initial.sae_reason.name), wrap_length)
        )
    except AttributeError:
        context["sae_reason"] = ""
    context["relevant_history"] = format_html(
        "{}",
        mark_safe(wrapx(escape_braces(ae_followup.relevant_history), wrap_length)),  # nosec B703, B308
    )
    context["ae_description"] = format_html(
        "{}",
        mark_safe(wrapx(escape_braces(ae_followup.ae_initial.ae_description), wrap_length)),  # nosec B703, B308
    )
    return context


@register.inclusion_tag(select_description_template("aesusar"), takes_context=True)
def format_ae_susar_description(context, ae_susar, wrap_length):
    context["utc_date"] = timezone.now().date()
    context["SHORT_DATE_FORMAT"] = settings.SHORT_DATE_FORMAT
    context["OTHER"] = OTHER
    context["YES"] = YES
    context["ae_susar"] = ae_susar
    context["ae_initial"] = ae_susar.ae_initial
    context["sae_reason"] = format_html(
        "{}",
        mark_safe(
            "<BR>".join(
                wrap(
                    escape_braces(ae_susar.ae_initial.sae_reason.name),
                    wrap_length or 35,
                )
            )
        ),  # nosec B703, B308
    )
    context["ae_description"] = format_html(
        "{}",
        mark_safe(wrapx(escape_braces(ae_susar.ae_initial.ae_description), wrap_length)),  # nosec B703, B308
    )
    return context


@register.inclusion_tag(
    "edc_adverse_event/tmg/tmg_ae_listboard_result.html",
    takes_context=True,
)
def tmg_listboard_results(
    context, results: [ActionItem], empty_qs_message: str | None = None
) -> dict:
    context["results"] = results
    context["empty_message"] = empty_qs_message
    return context


@register.inclusion_tag(
    "edc_adverse_event/tmg/death_report_tmg_panel.html",
    takes_context=True,
)
def render_death_report_tmg_panel(context, action_item: ActionItem = None):
    return dict(action_item=action_item)


@register.simple_tag
def ae_tmg_queryset(subject_identifier) -> QuerySet[DeathReportTmgModel]:
    return get_ae_model("aetmg").objects.filter(subject_identifier=subject_identifier)


@register.simple_tag
def death_report_tmg_queryset(subject_identifier: str) -> QuerySet[DeathReportTmgModel]:
    return get_ae_model("deathreporttmg").objects.filter(subject_identifier=subject_identifier)


@register.simple_tag
def death_report_tmg2_queryset(subject_identifier: str) -> QuerySet[DeathReportTmgSecondModel]:
    return get_ae_model("deathreporttmgsecond").objects.filter(
        subject_identifier=subject_identifier
    )


@register.simple_tag
def death_report_queryset(subject_identifier: str) -> QuerySet[DeathReportTmgSecondModel]:
    return get_ae_model("deathreport").objects.filter(subject_identifier=subject_identifier)


@register.simple_tag
def ae_followup_queryset(ae_initial: AeInitialModel = None) -> QuerySet[AeFollowupModel]:
    if ae_initial:
        return get_ae_model("aefollowup").objects.filter(ae_initial=ae_initial)
    return get_ae_model("aefollowup").objects.none()


@register.simple_tag
def ae_tmg_action_item_queryset(subject_identifier: str, *status) -> QuerySet[ActionItem]:
    return django_apps.get_model("edc_action_item.actionitem").objects.filter(
        subject_identifier=subject_identifier,
        action_type__name=AE_TMG_ACTION,
        status__in=status,
    )


@register.simple_tag
def death_report_tmg_action_item(subject_identifier: str) -> ActionItem:
    try:
        obj = django_apps.get_model("edc_action_item.actionitem").objects.get(
            subject_identifier=subject_identifier,
            action_type__name=DEATH_REPORT_TMG_ACTION,
        )
    except ObjectDoesNotExist:
        obj = None
    return obj


@register.simple_tag
def death_report_tmg_second_action_item(subject_identifier: str) -> ActionItem:
    try:
        obj = django_apps.get_model("edc_action_item.actionitem").objects.get(
            subject_identifier=subject_identifier,
            action_type__name=DEATH_REPORT_TMG_SECOND_ACTION,
        )
    except ObjectDoesNotExist:
        obj = None
    return obj


@register.inclusion_tag(
    "edc_adverse_event/tmg/ae_tmg_panel.html",
    takes_context=True,
)
def render_tmg_panel(
    context,
    action_item: ActionItem = None,
    reference_obj: DeathReportTmgModel = None,
    view_only: bool | None = None,
    by_user_created_only: bool | None = None,
    counter: int = None,
    report_status: str | None = None,
    next_url_name: str | None = None,
) -> dict:
    reference_obj = reference_obj or get_reference_obj(action_item)
    if not action_item and reference_obj:
        action_item = reference_obj.action_item
    disable_all = bool(not has_valid_tmg_perms(request=context["request"]))
    if action_item:
        params = dict(
            user=context["request"].user,
            subject_identifier=action_item.subject_identifier,
            model_obj=reference_obj,
            model_cls=action_item.action_cls.reference_model_cls(),
            request=context["request"],
            only_user_created_may_access=by_user_created_only,
            forloop_counter=counter,
            current_site=context["request"].site,
            disable_all=disable_all,
            action_item=action_item,
        )
        if next_url_name:
            params.update(next_url_name=next_url_name.split(":")[1])
        btn = TmgButton(**params)
        if view_only:
            panel_color = "info"
        elif not reference_obj:
            panel_color = "warning"
        else:
            panel_color = "success"
        # panel_label
        display_name = action_item.display_name.replace("Submit", "").replace("pending", "")
        identifier = action_item.identifier or "New"
        panel_label = _("%(display_name)s %(identifier)s") % dict(
            display_name=display_name, identifier=identifier
        )
        return dict(
            btn=btn,
            panel_color=panel_color,
            reference_obj=reference_obj,
            action_item=action_item,
            OPEN=OPEN,
            CLOSED=CLOSED,
            report_status=report_status,
            panel_label=panel_label,
        )
    return {}


@register.simple_tag(takes_context=True)
def has_perms_for_tmg_role(context):
    has_perms = False
    try:
        has_perms = context["request"].user.userprofile.roles.get(name=TMG_ROLE)
    except ObjectDoesNotExist:
        add_to_messages_once(
            context["request"],
            ERROR,
            _(
                "Access disabled. User has not been granted a TMG role. "
                "Contact your administrator."
            ),
        )
    return has_perms


@register.simple_tag()
def get_empty_qs_message(status: str, search_term: str):
    msg = f"There are no {status} reports."
    if search_term:
        msg = f"{msg[:-1]} for your search criteria"
    return _(msg)


@register.inclusion_tag(
    "edc_adverse_event/tmg/tmg_button_group.html",
    takes_context=True,
)
def render_tmg_button_group(context, subject_identifier: str):
    if context["request"].user.userprofile.roles.filter(name=TMG_ROLE).exists():
        return dict(subject_identifier=subject_identifier)
    return {}



from django import forms

from edc_form_validators import FormValidator, FormValidatorMixin

from ..models import QueryRule


class QueryRuleFormValidator(FormValidator):
    def clean(self):
        models = list(set([obj.model for obj in self.cleaned_data.get("data_dictionaries")]))
        if len(models) > 1:
            raise forms.ValidationError(
                {"data_dictionaries": "Invalid. Select questions from one CRF only"}
            )


class QueryRuleForm(FormValidatorMixin, forms.ModelForm):
    form_validator_cls = QueryRuleFormValidator

    class Meta:
        model = QueryRule
        fields = "__all__"



# Generated by Django 5.2.10 on 2026-01-25 17:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0129_historicalrepackrequest_override_container_unit_qty_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstocktransfer",
            name="cancel",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Leave blank. Otherwise type 'CANCEL' to cancel this transfer.",
                max_length=15,
                verbose_name="To cancel this transfer, type 'CANCEL'",
            ),
        ),
        migrations.AddField(
            model_name="historicalstocktransfer",
            name="comment",
            field=models.TextField(blank=True, default="", max_length=255),
        ),
        migrations.AddField(
            model_name="stocktransfer",
            name="cancel",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Leave blank. Otherwise type 'CANCEL' to cancel this transfer.",
                max_length=15,
                verbose_name="To cancel this transfer, type 'CANCEL'",
            ),
        ),
        migrations.AddField(
            model_name="stocktransfer",
            name="comment",
            field=models.TextField(blank=True, default="", max_length=255),
        ),
    ]



# Generated by Django 6.0 on 2025-01-28 13:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_facility", "0013_alter_healthfacility_site_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="healthfacility",
            name="title",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="historicalhealthfacility",
            name="title",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
    ]



from typing import Any

from django.core.handlers.wsgi import WSGIRequest

NAME = 0
MODEL_CLS = 1


class ModelAdminLimitToSelectedForeignkeyError(Exception):
    pass


class ModelAdminLimitToSelectedForeignkey:
    limit_fk_field_to_selected: list[tuple[str, Any]] = None

    def formfield_for_foreignkey(self, db_field, request: WSGIRequest, **kwargs):
        db = kwargs.get("using")
        if db_field.name in [item[NAME] for item in self.limit_fk_field_to_selected]:
            model_cls = None
            for item in self.limit_fk_field_to_selected:
                if item[NAME] == db_field.name:
                    model_cls = item[MODEL_CLS]
                    break
            if not model_cls:
                raise ModelAdminLimitToSelectedForeignkeyError(
                    f"Invalid model or model not found. Got {(db_field.name, model_cls)}"
                )
            if request.GET.get(db_field.name):
                kwargs["queryset"] = model_cls.on_site.using(db).filter(
                    id__exact=request.GET.get(db_field.name, 0)
                )
            else:
                kwargs["queryset"] = model_cls.objects.none()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)



from django.contrib import admin, messages
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext


@admin.display(description="Allocate stock to subject")
def allocate_stock_to_subject(modeladmin, request, queryset):
    """
    1. is there an open unprocess stock request?
    2. what stock is available at the site?
    3. what stock is available at central?

    """
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        url = reverse(
            "edc_pharmacy:allocate_url",
            args=(queryset.first().id,),
        )
        return HttpResponseRedirect(url)
    return None



from __future__ import annotations

import json

import mempass
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.core.handlers.wsgi import WSGIRequest
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.text import slugify
from django.views.generic.base import TemplateView

if getattr(settings, "EDC_PDF_REPORTS_INTEGRATE_EDC", True):
    from edc_dashboard.view_mixins import EdcViewMixin
    from edc_protocol.view_mixins import EdcProtocolViewMixin
else:

    class EdcViewMixin:
        pass

    class EdcProtocolViewMixin:
        pass


template_name = getattr(
    settings,
    "EDC_PDF_REPORTS_TEMPLATES",
    {"pdf_intermediate": "edc_pdf_reports/pdf_intermediate_edc.html"},
)


@method_decorator(login_required, name="dispatch")
class PdfIntermediateView(EdcViewMixin, EdcProtocolViewMixin, TemplateView):
    model_pks: list[str] | None = None
    template_name: str = template_name.get("pdf_intermediate")
    pdf_report_url_name = "edc_pdf_reports:pdf_report_url"
    session_key = "model_pks"
    mempass_nwords = 2

    def get(self, request: WSGIRequest, *args, **kwargs):
        if not self.model_pks:
            self.model_pks = [kwargs.get("pk")]
        request.session[self.session_key] = json.dumps([str(pk) for pk in self.model_pks])
        return super().get(request, *args, **kwargs)

    def get_context_data(self, app_label: str = None, model_name: str = None, **kwargs):
        model_cls = django_apps.get_model(app_label, model_name)
        kwargs.update(
            object_count=len(self.model_pks),
            report_name=model_cls._meta.verbose_name,
            url=self.get_pdf_report_url(app_label, model_name),
            return_to_changelist_url=self.return_to_changelist_url(model_cls),
            phrase=slugify(mempass.mkpassword(self.mempass_nwords)),
        )
        return super().get_context_data(**kwargs)

    def get_pdf_report_url(self, app_label: str, model_name: str) -> str:
        return reverse(
            self.pdf_report_url_name,
            kwargs=dict(app_label=app_label, model_name=model_name),
        )

    def return_to_changelist_url(self, model_cls):
        return reverse(model_cls.pdf_report_cls.changelist_url)



from __future__ import annotations

import re

from .constants import HIGH_VALUE
from .exceptions import ValueBoundryError


class InvalidUnits(Exception):  # noqa: N818
    pass


class InvalidLowerBound(Exception):  # noqa: N818
    pass


class InvalidLowerLimitNormal(Exception):  # noqa: N818
    pass


class InvalidUpperLimitNormal(Exception):  # noqa: N818
    pass


class InvalidUpperBound(Exception):  # noqa: N818
    pass


class InvalidCombination(Exception):  # noqa: N818
    pass


class Evaluator:
    def __init__(
        self,
        name: str | None = None,
        lower: int | float | None = None,
        upper: int | float | None = None,
        units: str | None = None,
        lower_inclusive: bool | None = None,
        upper_inclusive: bool | None = None,
        **kwargs,  # noqa: ARG002
    ) -> None:
        self.name = name
        if lower is not None and not re.match(r"\d+", str(lower)):
            raise InvalidLowerBound(f"Got {lower}.")
        if upper is not None and not re.match(r"\d+", str(upper)):
            raise InvalidUpperBound(f"Got {upper}.")
        self.lower: float | None = None if lower is None else float(lower)
        self.upper: float | None = None if upper is None else float(upper)

        if self.lower is not None and self.upper is not None:
            if self.lower == self.upper:
                raise InvalidCombination(
                    f"Lower and upper bound cannot be equal. Got {lower}={upper}"
                )
            if self.lower > self.upper:
                raise InvalidCombination(
                    f"Lower bound cannot exceed upper bound. Got {lower}>{upper}"
                )
        if not units:
            raise InvalidUnits("Got 'units' is None")
        self.units = units
        self.lower_inclusive = lower_inclusive
        self.upper_inclusive = upper_inclusive
        self.lower_operator: str | None = (
            None if self.lower is None else "<=" if self.lower_inclusive is True else "<"
        )
        self.upper_operator: str | None = (
            None if self.upper is None else "<=" if self.upper_inclusive is True else "<"
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.description()})"

    def __str__(self) -> str:
        return self.description()

    def description(
        self,
        value: int | float | None = None,
        show_as_int: bool | None = None,
        placeholder: str | None = None,
    ) -> str:
        placeholder = placeholder or "x"
        if show_as_int:
            value = int(value) if value is not None else placeholder
            lower = int(self.lower) if self.lower is not None else ""
            upper = int(self.upper) if self.upper is not None else ""
        else:
            value = float(value) if value is not None else placeholder
            lower = float(self.lower) if self.lower is not None else ""
            upper = float(self.upper) if self.upper is not None else ""
        if upper and upper >= float(HIGH_VALUE):
            upper = ""
        return (
            f"{lower}{self.lower_operator or ''}{value}"
            f"{self.upper_operator or ''}{upper} {self.units}"
        )

    def in_bounds_or_raise(self, value: int | float, units: str) -> bool:
        """Raises a ValueBoundryError exception if condition not met.

        The condition is evaluated to True or False as a string
        constructed from given parameters.

        For example,
            "lower lower_operator value upper_operator upper"
            "1.7<3.6<=3.5"
            "7.3<3.6"
        """
        value = float(value)
        if units != self.units:
            raise InvalidUnits(f"Expected {self.units}. See {self!r}")
        condition_str = (
            f"{'' if self.lower is None else self.lower}{self.lower_operator or ''}{value}"
            f"{self.upper_operator or ''}{'' if self.upper is None else self.upper}"
        )
        if not eval(condition_str):  # nosec B307  # noqa: S307
            raise ValueBoundryError(condition_str)
        return True



# Generated by Django 5.1.2 on 2024-11-17 20:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0037_remove_historicalstock_confirmed_at_site_by_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstocktransferconfirmation",
            name="transfer_confirmation_identifier",
            field=models.CharField(
                blank=True,
                db_index=True,
                help_text="A sequential unique identifier set by the EDC",
                max_length=36,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="stocktransferconfirmation",
            name="transfer_confirmation_identifier",
            field=models.CharField(
                blank=True,
                help_text="A sequential unique identifier set by the EDC",
                max_length=36,
                null=True,
                unique=True,
            ),
        ),
    ]



from django.apps import AppConfig as BaseAppConfig


class AppConfig(BaseAppConfig):
    name = "edc_reportable"
    verbose_name = "Edc Reportable"
    app_label = "edc_reportable"
    has_exportable_data = True
    include_in_administration_section = True



from django.contrib.sites.models import Site

from edc_sites.site import sites as site_sites
from edc_visit_schedule.models import VisitSchedule as BaseVisitSchedule


class VisitSchedule(BaseVisitSchedule):
    class Meta:
        proxy = True


class SiteProxy(Site):

    def __str__(self):
        single_site = site_sites.get(self.id)
        return single_site.description

    class Meta:
        proxy = True
        verbose_name = "Site"
        verbose_name_plural = "Sites"



# Generated by Django 2.2.6 on 2019-10-22 03:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_notification", "0004_auto_20190305_0123")]

    operations = [
        migrations.AddField(
            model_name="notification",
            name="mailing_list_address",
            field=models.EmailField(
                blank=True,
                max_length=254,
                null=True,
                verbose_name="Mailing list address",
            ),
        )
    ]



NOTIFICATION = "NOTIFICATION"
codenames = [
    "edc_notification.add_notification",
    "edc_notification.change_notification",
    "edc_notification.delete_notification",
    "edc_notification.view_notification",
]



import re

from django import forms

from edc_form_validators import FormValidator

from ..models import Box


class BoxForm(forms.ModelForm):
    def clean(self) -> dict:
        cleaned_data = super().clean()
        form_validator = FormValidator(cleaned_data=cleaned_data)
        if cleaned_data.get("specimen_types"):
            pattern = "([0-9][0-9]*[ ]*,[ ]*)*[0-9][0-9]*"
            match = re.match(pattern, cleaned_data.get("specimen_types"))
            if not match or match.group() != cleaned_data.get("specimen_types"):
                raise forms.ValidationError(
                    {"specimen_types": "Invalid list of specimen types."},
                    code="invalid format",
                )
            specimen_types = [code.strip() for code in match.group().split(",")]
            if len(specimen_types) != len(list(set(specimen_types))):
                raise forms.ValidationError(
                    {"specimen_types": "List must be unique."},
                    code="list not unique",
                )
            cleaned_data["specimen_types"] = ",".join(specimen_types)
        form_validator.validate_other_specify("category")
        return cleaned_data

    class Meta:
        fields = "__all__"
        model = Box



from datetime import datetime
from typing import Any, Protocol, TypeVar
from uuid import UUID

from django.db import models

from edc_visit_schedule.schedule import Schedule


class AppointmentModelStub(Protocol):
    id: UUID | models.UUIDField
    pk: UUID | models.UUIDField
    subject_identifier: str | models.CharField
    appt_datetime: datetime | models.DateTimeField
    visit_code: str | models.CharField
    visit_code_sequence: int | models.IntegerField
    visit_schedule_name: str | models.CharField
    schedule_name: str | models.CharField
    facility_name: str | models.CharField
    timepoint: int | models.IntegerField
    timepoint_datetime: datetime
    schedule: Schedule
    _meta: Any

    objects: models.Manager

    last_visit_code_sequence: int | None
    next: "AppointmentModelStub"
    previous: "AppointmentModelStub"
    get_next: "AppointmentModelStub"

    def save(self, *args, **kwargs) -> None: ...

    def natural_key(self) -> tuple: ...

    def get_previous(self) -> "AppointmentModelStub": ...

    @classmethod
    def related_visit_model_attr(cls) -> str: ...


TAppointmentModelStub = TypeVar("TAppointmentModelStub", bound="AppointmentModelStub")



from .allocation import Allocation

__all__ = ["AllocationProxy"]


class AllocationProxy(Allocation):
    class Meta:
        proxy = True
        verbose_name = "Allocation"
        verbose_name_plural = "Allocation"



from edc_randomization.auth_objects import RANDO_UNBLINDED
from edc_randomization.blinding import user_is_blinded


def blinded_user(request) -> bool:
    if user_is_blinded(request.user.username) or (
        not user_is_blinded(request.user.username)
        and RANDO_UNBLINDED not in [g.name for g in request.user.groups.all()]
    ):
        # user is blinded
        return True
    # user is not blinded
    return False


__all__ = ["blinded_user"]



from django.contrib import admin
from edc_model.admin import HistoricalModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils import to_local
from rangefilter.filters import DateRangeFilterBuilder

from ....admin_site import edc_pharmacy_history_admin
from ....models import Confirmation
from ...model_admin_mixin import ModelAdminMixin


@admin.register(Confirmation.history.model, site=edc_pharmacy_history_admin)
class ConfirmationHistoryAdmin(HistoricalModelAdminMixin, ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Confirmation of created stock (History)"
    change_form_title = "Pharmacy: onfirmation of created stock (History)"
    change_list_note_url_name = "edc_pharmacy_admin:edc_pharmacy_confirmation_changelist"

    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    list_display = (
        "item_id",
        "item_date",
        "history_type",
        "formatted_history_date",
        "code",
        "revision",
        "history_id",
    )

    list_filter = (
        "history_type",
        ("history_date", DateRangeFilterBuilder()),
    )

    search_fields = (
        "code",
        "history_id",
        "confirmation_identifier",
    )

    @admin.display(description="ITEM#", ordering="confirmation_identifier")
    def item_id(self, obj):
        return obj.confirmation_identifier

    @admin.display(description="Date", ordering="confirmed_datetime")
    def item_date(self, obj):
        if obj.confirmed_datetime:
            return to_local(obj.confirmed_datetime).date()
        return None



from edc_navbar import Navbar, NavbarItem, site_navbars

navbar = Navbar(name="edc_visit_schedule")

navbar.register(
    NavbarItem(
        name="visit_schedule",
        title="Visit Schedule",
        label="Visit Schedule",
        fa_icon="fa-calendar",
        url_with_namespace="edc_visit_schedule:home_url",
        codename="edc_navbar.nav_visit_schedule",
    )
)

navbar.register(
    NavbarItem(
        name="admin",
        title="Subject History",
        label="Subject History",
        fa_icon="fa-history",
        url_with_namespace=(
            "edc_visit_schedule_admin:edc_visit_schedule_subjectschedulehistory_changelist"
        ),
        codename="edc_navbar.nav_visit_schedule",
    )
)

site_navbars.register(navbar)



import urllib
from typing import Any


class QueryStringViewMixin:
    @property
    def querystring(self):
        querystring = {}
        f = self.request.GET.get("f")
        e = self.request.GET.get("e")
        q = self.request.GET.get("q")
        o = self.request.GET.get("o")
        if f:
            querystring.update(f=f)
        if e:
            querystring.update(e=e)
        if q:
            querystring.update(q=q)
        if o:
            querystring.update(o=o)
        if querystring:
            return "?" + urllib.parse.urlencode(querystring)
        return ""

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            f=self.request.GET.get("f"),
            e=self.request.GET.get("e"),
            o=self.request.GET.get("o"),
            q=self.request.GET.get("q"),
            querystring=self.querystring,
        )
        return super().get_context_data(**kwargs)



from uuid import UUID

from django.apps import apps as django_apps


def create_new_stock_on_receive(receive_item_pk: UUID = None):
    receive_item_model_cls = django_apps.get_model("edc_pharmacy.receiveitem")
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    receive_item = receive_item_model_cls.objects.get(pk=receive_item_pk)
    if (
        create_count := (
            receive_item.item_qty_received
            - stock_model_cls.objects.filter(receive_item_id=receive_item.id).count()
        )
        > 0
    ):
        for _ in range(0, create_count):
            stock_model_cls.objects.create(
                receive_item_id=receive_item.id,
                qty_in=1,
                qty_out=0,
                qty=1,
                product_id=receive_item.order_item.product.id,
                container_id=receive_item.container.id,
                location_id=receive_item.receive.location.id,
                lot_id=receive_item.lot.id,
            )



from django.contrib import admin
from django.contrib.admin.decorators import register

from edc_model_admin.mixins import TemplatesModelAdminMixin

from .admin_site import edc_notification_admin
from .models import Notification


@register(Notification, site=edc_notification_admin)
class NotificationAdmin(TemplatesModelAdminMixin, admin.ModelAdmin):
    model = Notification

    list_display = ("name", "display_name", "mailing_list_address")



from django.db import models


class OnStudyMissingValuesModelMixin(models.Model):
    original_id = models.UUIDField(null=True)

    label_lower = models.CharField(max_length=150, default="")

    subject_visit_id = models.UUIDField(null=True)

    report_datetime = models.DateTimeField(null=True)

    label = models.CharField(max_length=50, default="")

    visit_code = models.CharField(max_length=25, default="")

    visit_code_sequence = models.IntegerField(null=True)

    schedule_name = models.CharField(max_length=25, default="")

    modified = models.DateTimeField(null=True)

    class Meta:
        abstract = True



# Generated by Django 4.1.2 on 2022-11-30 22:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_protocol_incident",
            "0016_historicalprotocolincident_reasons_withdrawn_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="reasons_withdrawn",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="reasons_withdrawn",
            field=models.TextField(blank=True, null=True),
        ),
    ]



from dateutil.relativedelta import relativedelta
from django.utils import timezone
from django.utils.timezone import localtime

from edc_reportable.age_evaluator import AgeEvaluator as ReportableAgeEvaluator
from edc_reportable.exceptions import ValueBoundryError


class AgeEvaluator(ReportableAgeEvaluator):
    def __init__(self, **kwargs) -> None:
        self.reasons_ineligible: str = ""
        super().__init__(**kwargs)

    def eligible(self, age: int | None = None) -> bool:
        self.reasons_ineligible = ""
        eligible = False
        if age:
            try:
                self.in_bounds_or_raise(age=age)
            except ValueBoundryError as e:
                self.reasons_ineligible = str(e)
            else:
                eligible = True
        else:
            self.reasons_ineligible = "Age unknown"
        return eligible

    def in_bounds_or_raise(self, age: int | None = None, **kwargs):  # noqa: ARG002
        self.reasons_ineligible = ""
        dob = localtime(timezone.now() - relativedelta(years=age)).date()
        age_units = "years"
        report_datetime = localtime(timezone.now())
        return super().in_bounds_or_raise(
            dob=dob, report_datetime=report_datetime, age_units=age_units
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from clinicedc_constants import CLOSED, NEW, OPEN
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_dashboard.url_names import url_names
from edc_dashboard.view_mixins import EdcViewMixin
from edc_listboard.view_mixins import ListboardFilterViewMixin, SearchFormViewMixin
from edc_listboard.views import ListboardView as BaseListboardView
from edc_navbar import NavbarViewMixin

from ...constants import AE_INITIAL_ACTION
from ...pdf_reports import AePdfReport
from ...utils import get_ae_model

if TYPE_CHECKING:
    from django.db.models import Q, QuerySet


class AeListboardViewMixin(
    NavbarViewMixin,
    EdcViewMixin,
    ListboardFilterViewMixin,
    SearchFormViewMixin,
    BaseListboardView,
):
    pdf_report_cls = AePdfReport

    listboard_back_url = "ae_home_url"
    home_url = "ae_home_url"
    listboard_panel_title = "Adverse Events: AE Initial and Follow-up Reports"

    listboard_template = "ae_listboard_template"
    listboard_url = "ae_listboard_url"
    listboard_panel_style = "default"
    listboard_model = "edc_action_item.actionitem"
    listboard_view_permission_codename = "edc_adverse_event.view_ae_listboard"

    listboard_instructions = format_html(
        "{}",
        mark_safe(render_to_string("edc_adverse_event/ae/ae_listboard_instructions.html")),  # nosec B703 B308,
    )
    navbar_selected_item = "ae_home"
    ordering = "-report_datetime"
    paginate_by = 25
    search_form_url = "ae_listboard_url"
    action_type_names = (AE_INITIAL_ACTION,)

    search_fields = (
        "subject_identifier",
        "action_identifier",
        "parent_action_item__action_identifier",
        "related_action_item__action_identifier",
        "user_created",
        "user_modified",
    )

    @property
    def ae_initial_model_cls(self):
        return get_ae_model("aeinitial")

    def get(self, request, *args, **kwargs):
        response = None
        if request.GET.get("pdf"):
            response = self.print_pdf_report(
                action_identifier=self.request.GET.get("pdf"),
                request=request,
            )
        return response or super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            AE_INITIAL_ACTION=AE_INITIAL_ACTION,
            utc_date=timezone.now().date(),
            **{"ae_home_url": url_names.get(self.home_url)},
        )
        return super().get_context_data(**kwargs)

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        options.update(
            action_type__name__in=self.action_type_names,
            status__in=[NEW, OPEN, CLOSED],
        )
        if kwargs.get("subject_identifier"):
            options.update({"subject_identifier": kwargs.get("subject_identifier")})
        return q_object, options

    def get_updated_queryset(self, queryset) -> QuerySet:
        pks = []
        for obj in queryset:
            try:
                obj.reference_obj  # noqa: B018
            except ObjectDoesNotExist:
                pks.append(obj.pk)
        return queryset.exclude(pk__in=pks)

    def print_pdf_report(self, action_identifier=None, request=None):
        try:
            ae_initial_obj = self.ae_initial_model_cls.objects.get(
                action_identifier=action_identifier
            )
        except ObjectDoesNotExist:
            pass
        else:
            pdf_report = self.get_pdf_report(
                ae_initial=ae_initial_obj,
                subject_identifier=ae_initial_obj.subject_identifier,
                user=self.request.user,
                request=request,
            )
            return pdf_report.render_to_response()
        return None

    def get_pdf_report(self, **kwargs) -> AePdfReport:
        pdf_report_cls = getattr(self.ae_initial_model_cls, "pdf_report_cls", None)
        if not pdf_report_cls:
            pdf_report_cls = getattr(self, "pdf_report_cls", None)
        return pdf_report_cls(**kwargs)



from .applicable_field_validator import ApplicableFieldValidator
from .base_form_validator import BaseFormValidator
from .current_site_validator import CurrentSiteValidator
from .date_range_validator import DateRangeFieldValidator
from .date_validator import DateValidator
from .locale_validator import LocaleValidator
from .many_to_many_field_validator import ManyToManyFieldValidator
from .other_specify_field_validator import OtherSpecifyFieldValidator
from .range_field_validator import RangeFieldValidator
from .required_field_validator import RequiredFieldValidator


class FormValidator(
    RequiredFieldValidator,
    ManyToManyFieldValidator,
    OtherSpecifyFieldValidator,
    ApplicableFieldValidator,
    RangeFieldValidator,
    DateRangeFieldValidator,
    DateValidator,
    CurrentSiteValidator,
    LocaleValidator,
    BaseFormValidator,
):
    pass



# Generated by Django 6.0 on 2026-01-18 01:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0098_remove_historicalstock_returned_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstock",
            name="confirmed_at_location",
            field=models.BooleanField(
                default=False, help_text="See confirmeatlocationitem."
            ),
        ),
        migrations.AddField(
            model_name="stock",
            name="confirmed_at_location",
            field=models.BooleanField(
                default=False, help_text="See confirmeatlocationitem."
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="in_transit",
            field=models.BooleanField(
                default=False, help_text="See stocktransferitem."
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="in_transit",
            field=models.BooleanField(
                default=False, help_text="See stocktransferitem."
            ),
        ),
    ]



from clinicedc_constants import NULL_STRING, TBD
from django.db import models
from django.utils import timezone

from edc_action_item.models.action_model_mixin import ActionModelMixin
from edc_constants.choices import YES_NO_TBD
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models.base_uuid_model import BaseUuidModel
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin

from ..constants import UNBLINDING_REVIEW_ACTION
from .unblinding_user import UnblindingReviewerUser


class UnblindingReview(
    NonUniqueSubjectIdentifierFieldMixin,
    SiteModelMixin,
    ActionModelMixin,
    BaseUuidModel,
):
    action_name = UNBLINDING_REVIEW_ACTION

    report_datetime = models.DateTimeField(
        verbose_name="Report Date and Time", default=timezone.now
    )

    reviewer = models.ForeignKey(
        UnblindingReviewerUser,
        related_name="+",
        on_delete=models.PROTECT,
        verbose_name="Unblinding request reviewed by",
        help_text="Choose a name from the list",
    )

    approved = models.CharField(max_length=15, default=TBD, choices=YES_NO_TBD)

    comment = models.TextField(verbose_name="Comment", default=NULL_STRING)

    objects = SubjectIdentifierManager()

    on_site = CurrentSiteManager()

    def natural_key(self):
        return (self.action_identifier,)

    class Meta(BaseUuidModel.Meta, NonUniqueSubjectIdentifierFieldMixin.Meta):
        verbose_name = "Unblinding Review"
        verbose_name_plural = "Unblinding Reviews"
        indexes = (
            models.Index(fields=["subject_identifier", "action_identifier", "site", "id"]),
        )



from edc_listboard.filters import ListboardFilter, ListboardViewFilters


class ManifestListboardViewFilters(ListboardViewFilters):
    all = ListboardFilter(name="all", label="All", lookup={})

    shipped = ListboardFilter(label="Shipped", lookup={"shipped": True})

    not_shipped = ListboardFilter(label="Not shipped", lookup={"shipped": False})

    printed = ListboardFilter(label="Printed", lookup={"printed": True})

    not_printed = ListboardFilter(label="Not printed", default=True, lookup={"printed": False})



from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from edc_model.validators import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_visit_schedule.model_mixins import CrfScheduleModelMixin

from ..base import VisitMethodsModelMixin


class VisitTrackingRequisitionModelMixin(
    VisitMethodsModelMixin,
    CrfScheduleModelMixin,
    models.Model,
):
    """Model mixin used by RequisitionModelMixin (edc-lab)"""

    subject_visit = models.ForeignKey(settings.SUBJECT_VISIT_MODEL, on_delete=models.PROTECT)

    report_datetime = models.DateTimeField(
        verbose_name=_("Report Date"),
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=timezone.now,
        help_text=_(
            "If reporting today, use today's date/time, otherwise use "
            "the date/time this information was reported."
        ),
    )

    @classmethod
    def related_visit_model_attr(cls) -> str:
        # assuming subject_visit
        return "subject_visit"

    class Meta:
        abstract = True



from django.urls.conf import path, re_path

from edc_dashboard.url_names import url_names

from .dashboard_urls import dashboard_urls
from .views import (
    AliquotListboardView,
    AliquotView,
    HomeView,
    ManageBoxItemView,
    ManageBoxListboardView,
    ManageManifestListboardView,
    ManageManifestView,
    ManifestListboardView,
    ManifestView,
    PackListboardView,
    PackView,
    ProcessListboardView,
    ProcessView,
    ReceiveListboardView,
    ReceiveView,
    RequisitionListboardView,
    RequisitionView,
    ResultListboardView,
    VerifyBoxItemView,
    VerifyBoxListboardView,
)

app_name = "edc_lab_dashboard"

urlpatterns = [
    # listboard urls
    re_path(
        "listboard/requisition/(?P<page>[0-9]+)/$",
        RequisitionListboardView.as_view(),
        name="requisition_listboard_url",
    ),
    path(
        "listboard/requisition/",
        RequisitionListboardView.as_view(),
        name="requisition_listboard_url",
    ),
    re_path(
        "listboard/receive/(?P<page>[0-9]+)/$",
        ReceiveListboardView.as_view(),
        name="receive_listboard_url",
    ),
    path(
        "listboard/receive/",
        ReceiveListboardView.as_view(),
        name="receive_listboard_url",
    ),
    re_path(
        "listboard/process/(?P<page>[0-9]+)/$",
        ProcessListboardView.as_view(),
        name="process_listboard_url",
    ),
    path(
        "listboard/process/",
        ProcessListboardView.as_view(),
        name="process_listboard_url",
    ),
    re_path(
        "listboard/pack/(?P<page>[0-9]+)/$",
        PackListboardView.as_view(),
        name="pack_listboard_url",
    ),
    path("listboard/pack/", PackListboardView.as_view(), name="pack_listboard_url"),
    re_path(
        "listboard/box/(?P<action_name>manage)/"
        "(?P<box_identifier>[A-Z0-9]+)/(?P<page>[0-9]+)/$",
        ManageBoxListboardView.as_view(),
        name="manage_box_listboard_url",
    ),
    re_path(
        "listboard/box/(?P<action_name>manage)/(?P<box_identifier>[A-Z0-9]+)/$",
        ManageBoxListboardView.as_view(),
        name="manage_box_listboard_url",
    ),
    re_path(
        "listboard/box/(?P<action_name>manage)/",
        ManageBoxListboardView.as_view(),
        name="manage_box_listboard_url",
    ),
    re_path(
        "listboard/box/(?P<action_name>verify)/"
        "(?P<box_identifier>[A-Z0-9]+)/"
        "(?P<position>[0-9]+)/(?P<page>[0-9]+)/$",
        VerifyBoxListboardView.as_view(),
        name="verify_box_listboard_url",
    ),
    re_path(
        "listboard/box/(?P<action_name>verify)/"
        "(?P<box_identifier>[A-Z0-9]+)/"
        "(?P<position>[0-9]+)/$",
        VerifyBoxListboardView.as_view(),
        name="verify_box_listboard_url",
    ),
    re_path(
        "listboard/box/(?P<action_name>verify)/(?P<position>[0-9]+)/$",
        VerifyBoxListboardView.as_view(),
        name="verify_box_listboard_url",
    ),
    re_path(
        "listboard/aliquot/(?P<page>[0-9]+)/$",
        AliquotListboardView.as_view(),
        name="aliquot_listboard_url",
    ),
    path(
        "listboard/aliquot/",
        AliquotListboardView.as_view(),
        name="aliquot_listboard_url",
    ),
    re_path(
        "listboard/manifest/(?P<action_name>manage)/"
        "(?P<manifest_identifier>[A-Z0-9]+)/(?P<page>[0-9]+)/$",
        ManageManifestListboardView.as_view(),
        name="manage_manifest_listboard_url",
    ),
    re_path(
        "listboard/manifest/(?P<action_name>manage)/(?P<manifest_identifier>[A-Z0-9]+)/$",
        ManageManifestListboardView.as_view(),
        name="manage_manifest_listboard_url",
    ),
    re_path(
        "listboard/manifest/(?P<action_name>manage)/$",
        ManageManifestListboardView.as_view(),
        name="manage_manifest_listboard_url",
    ),
    re_path(
        "listboard/manifest/(?P<page>[0-9]+)/$",
        ManifestListboardView.as_view(),
        name="manifest_listboard_url",
    ),
    path(
        "listboard/manifest/",
        ManifestListboardView.as_view(),
        name="manifest_listboard_url",
    ),
    re_path(
        "listboard/result/(?P<page>[0-9]+)/$",
        ResultListboardView.as_view(navbar_selected_item="result"),
        name="result_listboard_url",
    ),
    path(
        "listboard/result/",
        ResultListboardView.as_view(navbar_selected_item="result"),
        name="result_listboard_url",
    ),
    # action urls
    path("requisition/", RequisitionView.as_view(), name="requisition_form_action_url"),
    path("requisition/receive/", ReceiveView.as_view(), name="receive_form_action_url"),
    path("requisition/process/", ProcessView.as_view(), name="process_form_action_url"),
    path("requisition/pack/", PackView.as_view(), name="pack_form_action_url"),
    re_path(
        "box/(?P<box_identifier>[A-Z0-9]+)/(?P<action_name>manage)/$",
        ManageBoxItemView.as_view(),
        name="manage_box_item_form_action_url",
    ),
    re_path(
        "box/(?P<box_identifier>[A-Z0-9]+)/(?P<action_name>verify)/(?P<position>[0-9]+)/$",
        VerifyBoxItemView.as_view(),
        name="verify_box_item_form_action_url",
    ),
    path("manifest/", ManifestView.as_view(), name="manifest_form_action_url"),
    re_path(
        "manifest/(?P<manifest_identifier>[A-Z0-9]+)/(?P<action_name>manage)/$",
        ManageManifestView.as_view(),
        name="manage_manifest_item_form_action_url",
    ),
    path("aliquot/", AliquotView.as_view(), name="aliquot_form_action_url"),
    path("", HomeView.as_view(), name="home_url"),
]

url_names.register_from_dict(**dashboard_urls)



from clinicedc_constants import COMPLETE, NEW, NOT_APPLICABLE, OTHER
from django.utils.translation import gettext as _

from .constants import (
    ALLOCATED,
    AVAILABLE,
    CANCELLED,
    DISPENSED,
    FILLED,
    PARTIAL,
    ZERO_ITEM,
)

PRESCRIPTION_STATUS = (
    (NEW, "New"),
    (PARTIAL, "Partially filled"),
    (FILLED, "Filled"),
    (CANCELLED, "Cancelled"),
)


DISPENSE_STATUS = ((DISPENSED, "Dispensed"), (CANCELLED, "Cancelled"))


FREQUENCY = (
    ("hr", "per hour"),
    ("day", "per day"),
    ("single", "single dose"),
    (OTHER, "Other ..."),
    (NOT_APPLICABLE, "Not applicable"),
)

ORDER_CHOICES = ((NEW, _("New")), (PARTIAL, _("Partial")), (COMPLETE, _("Complete")))

STOCK_STATUS = (
    (AVAILABLE, "Available"),
    (ALLOCATED, "Allocated"),
    (ZERO_ITEM, "Zero"),
)


STOCK_UPDATE = (
    ("edc_pharmacy.receiveitem", "Receiving"),
    ("edc_pharmacy.repackrequest", "Repacking"),
)



from .allocation_history_admin import AllocationHistoryAdmin
from .confirmation_at_location_history_admin import ConfirmationAtLocationHistoryAdmin
from .confirmation_at_location_item_history_admin import ConfirmationAtLocationItemHistoryAdmin
from .confirmation_history_admin import ConfirmationHistoryAdmin
from .dispense_item_history_admin import DispenseItemHistoryAdmin
from .stock_history_admin import StockHistoryAdmin
from .storage_bin_item_history_admin import StorageBinItemHistoryAdmin



from __future__ import annotations

from typing import Any

from clinicedc_constants import YES
from django.conf import settings
from django.contrib import messages
from django.db.models import Q
from django.utils.html import format_html

from ..listboard_filters import RequisitionListboardViewFilters
from .base_listboard_view import BaseListboardView


class RequisitionListboardView(BaseListboardView):
    listboard_model = settings.SUBJECT_REQUISITION_MODEL

    form_action_url = "requisition_form_action_url"
    listboard_template = "requisition_listboard_template"
    listboard_url = "requisition_listboard_url"
    listboard_view_filters = RequisitionListboardViewFilters()
    listboard_view_permission_codename = "edc_lab_dashboard.view_lab_requisition_listboard"
    listboard_view_only_my_permission_codename = None
    navbar_selected_item = "requisition"
    search_form_url = "requisition_listboard_url"
    show_all = True
    ordering = ["-modified", "-created"]

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        unverified_requisition_count = (
            self.get_queryset().filter(clinic_verified__isnull=True).count()
        )
        if unverified_requisition_count:
            verb = "is" if unverified_requisition_count == 1 else "are"
            plural = "" if unverified_requisition_count == 1 else "s"
            messages.warning(
                self.request,
                format_html(
                    "There {} {} requisition{} "
                    "where the specimen is <b>drawn but not verified</b> by the clinic. "
                    "Please follow up.",
                    verb,
                    str(unverified_requisition_count),
                    plural,
                ),
            )
            kwargs.update(unverified_requisition_count=unverified_requisition_count)
        return super().get_context_data(**kwargs)

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_objects, options = super().get_queryset_filter_options(request, *args, **kwargs)
        options.update(is_drawn=YES)
        return Q(), options



from .aliquot_identifier_model_mixin import AliquotIdentifierModelMixin
from .aliquot_label_mixin import AliquotLabelMixin
from .aliquot_model_mixin import AliquotModelMixin
from .aliquot_shipping_mixin import AliquotShippingMixin
from .aliquot_type_model_mixin import AliquotTypeModelMixin



"""
https://ctep.cancer.gov/protocoldevelopment/electronic_applications/docs/CTCAE_v5_Quick_Reference_8.5x11.pdf
Gamma GT
U/L

Male: 12 <= x <= 64
G1 64 < x <= 160
G2 160 < x <= 320
G3 320 < x <= 1280
G4 x > 1280

Female: 9 - 36
>36 -≤90
>90 -≤180
>180 -≤720
>720

>ULN - 2.5 x ULN if baseline was normal; 2.0 - 2.5 x baseline if baseline was abnormal
>2.5 - 5.0 x ULN if baseline was normal; >2.5 - 5.0 x baseline if baseline was abnormal
>5.0 - 20.0 x ULN if baseline was normal; >5.0 - 20.0 x baseline if baseline was abnormal
>20.0 x ULN if baseline was normal; >20.0 x baseline if baseline was abnormal
"""

from clinicedc_constants import FEMALE, IU_LITER, MALE

from ...adult_age_options import adult_age_options
from ...formula import Formula

ggt_baseline_normal = (
    [
        Formula(
            "2.50*ULN<x<=2.50*ULN",
            grade=1,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2.50*ULN<=x<5.00*ULN",
            grade=2,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.00*ULN<=x<10.00*ULN",
            grade=3,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "10.00*ULN<=x",
            grade=4,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
)



from edc_auth.constants import EVERYONE
from edc_auth.site_auths import site_auths


def update_site_auths() -> None:
    site_auths.update_group("edc_visit_schedule.view_subjectschedulehistory", name=EVERYONE)


update_site_auths()



from django.http import HttpResponse
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import cm

from ..models import StockTransfer
from ..pdf_reports import ManifestReport, NumberedCanvas


def print_stock_transfer_manifest_view(request, stock_transfer: StockTransfer | None = None):

    stock_transfer = StockTransfer.objects.get(pk=stock_transfer)
    response = HttpResponse(content_type="application/pdf")
    response["Content-Disposition"] = (
        f'attachment; filename="stock_transfer_{stock_transfer.transfer_identifier}.pdf"'
    )
    page = dict(
        rightMargin=1 * cm,
        leftMargin=1 * cm,
        topMargin=1 * cm,
        bottomMargin=1.5 * cm,
        pagesize=A4,
    )
    report = ManifestReport(
        stock_transfer=stock_transfer,
        request=request,
        footer_row_height=60,
        page=page,
        numbered_canvas=NumberedCanvas,
    )

    report.build(response)
    return response



# Generated by Django 6.0 on 2026-01-19 16:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0102_dispenseitem_code_historicaldispenseitem_code_and_more"),
    ]

    operations = []



import contextlib
from collections import namedtuple
from typing import Any

from clinicedc_constants import NO, NULL_STRING, YES

from edc_lab import RequisitionPanel
from edc_lab.form_validators import CrfRequisitionFormValidatorMixin
from edc_reportable.forms import ReportablesFormValidatorMixin


class BloodResultsFormValidatorError(Exception):
    pass


class BloodResultsFormValidatorMixin(
    ReportablesFormValidatorMixin,
    CrfRequisitionFormValidatorMixin,
):
    value_field_suffix = "_value"
    panel: RequisitionPanel = None
    panels: tuple[RequisitionPanel, ...] = ()
    is_poc_field: str = "is_poc"

    def evaluate_value(self, **kwargs):
        """A hook to evaluate a field value"""
        pass

    def clean(self: Any) -> None:
        if self.cleaned_data.get(self.is_poc_field) not in [None, NULL_STRING]:
            # do not require requisition if poc == YES
            self.required_if(NO, field=self.is_poc_field, field_required="requisition")
        else:
            # requires requisition if any `value` fields have value but inverse not True.
            # It is OK to submit the requisition without any `value` fields data.
            self.required_if_true(
                any(self.fields_names_with_values),
                field_required=self.requisition_field,
                inverse=False,
            )

        if self.requisition:
            for fields_name in self.fields_names_with_values:
                try:
                    utest_id, _ = fields_name.split(self.value_field_suffix)
                except ValueError:
                    utest_id = fields_name
                if f"{utest_id}_units" in self.cleaned_data:
                    self.required_if_not_none(
                        field=f"{utest_id}{self.value_field_suffix or NULL_STRING}",
                        field_required=f"{utest_id}_units",
                        field_required_evaluate_as_int=True,
                    )
                if f"{utest_id}_abnormal" in self.cleaned_data:
                    self.required_if_not_none(
                        field=f"{utest_id}{self.value_field_suffix or NULL_STRING}",
                        field_required=f"{utest_id}_abnormal",
                        field_required_evaluate_as_int=False,
                    )
                if f"{utest_id}_reportable" in self.cleaned_data:
                    self.required_if_not_none(
                        field=f"{utest_id}{self.value_field_suffix or NULL_STRING}",
                        field_required=f"{utest_id}_reportable",
                        field_required_evaluate_as_int=False,
                    )
                self.evaluate_value(prefix=utest_id)
            self.validate_reportable_fields(
                self.requisition.panel_object.reference_range_collection_name,
                **self.reportables_evaluator_options,
            )
        super().clean()

    @property
    def requisition(self: Any):
        """Returns a Requisition instance or raises
        forms ValidationError.

        If POC, Requisition instance is not a model
        """
        if self.cleaned_data.get(self.is_poc_field) == YES:
            Requisition = namedtuple("Requisition", "panel_object")
            return Requisition(self.panel)
        return self.validate_requisition(*self.panel_list)

    @property
    def is_poc(self: Any) -> bool:
        if self.is_poc_field and self.cleaned_data.get(self.is_poc_field):
            return self.cleaned_data.get(self.is_poc_field) == YES
        return False

    @property
    def fields_names_with_values(self: Any) -> tuple[str, ...]:
        """Returns a list result `value` field names that are not None."""
        field_names = (f"{utest_id}{self.value_field_suffix}" for utest_id in self.utest_ids)
        return tuple(
            [
                field_name
                for field_name in field_names
                if self.cleaned_data.get(field_name) not in [None, NULL_STRING]
            ]
        )

    @property
    def utest_ids(self: Any) -> tuple:
        utest_ids = []
        for panel in self.panel_list:
            for utest_id in panel.utest_ids:
                with contextlib.suppress(ValueError):
                    utest_id, _ = utest_id  # noqa: PLW2901
                utest_ids.append(utest_id)
        return tuple(utest_ids)

    @property
    def panel_list(self: Any) -> tuple[RequisitionPanel]:
        if self.panel:
            return (self.panel,)
        return self.panels



# Generated by Django 5.2.10 on 2026-01-24 16:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0111_alter_container_qty_default_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstock",
            name="container_qty",
            field=models.DecimalField(decimal_places=2, max_digits=20, null=True),
        ),
        migrations.AddField(
            model_name="stock",
            name="container_qty",
            field=models.DecimalField(decimal_places=2, max_digits=20, null=True),
        ),
    ]



from .convert_dates_from_model import convert_dates_from_model
from .convert_numbers_to_float64_int64 import convert_numbers_to_float64_int64
from .convert_numbers_to_nullable_dtype import convert_numbers_to_nullable_dtype
from .convert_numerics_from_model import convert_numerics_from_model
from .convert_timedelta_from_model import convert_timedelta_from_model
from .convert_visit_code_to_float import convert_visit_code_to_float
from .get_export_folder import get_export_folder
from .model_from_table_name import get_model_from_table_name



# Generated by Django 2.2.2 on 2019-06-26 14:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_locator", "0018_auto_20190305_0123")]

    operations = [
        migrations.AddField(
            model_name="historicalsubjectlocator",
            name="action_item_reason",
            field=models.TextField(editable=False, null=True),
        ),
        migrations.AddField(
            model_name="subjectlocator",
            name="action_item_reason",
            field=models.TextField(editable=False, null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2022-08-25 01:51

import _socket
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django.utils.timezone
import django_revision.revision_field
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0029_auto_20220704_1841"),
    ]

    operations = [
        migrations.CreateModel(
            name="EdcPermissions",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
            ],
            options={
                "verbose_name": "Edc Permissions",
                "verbose_name_plural": "Edc Permissions",
                "abstract": False,
            },
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings

from .default_definitions import default_definitions
from .facility import Facility, FacilityError

if TYPE_CHECKING:
    from .models import HealthFacility, Holiday


def get_holiday_model() -> str:
    return getattr(settings, "EDC_FACILITY_HOLIDAY_MODEL", "edc_facility.holiday")


def get_holiday_model_cls() -> type[Holiday]:
    return django_apps.get_model(get_holiday_model())


def get_default_facility_name() -> str:
    return getattr(settings, "EDC_FACILITY_DEFAULT_FACILITY_NAME", "default")


def get_facility_definitions():
    return getattr(settings, "EDC_FACILITY_DEFINITIONS", default_definitions)


def get_facilities() -> dict[str, Facility]:
    """Returns a dictionary of facilities."""
    return {k: Facility(name=k, **v) for k, v in get_facility_definitions().items()}


def get_facility(name: str = None) -> Facility:
    """Returns a facility instance for this name, if it exists,
    or raises.
    """
    facilities = get_facilities()
    facility = facilities.get(name)
    if not facility:
        raise FacilityError(f"Facility '{name}' does not exist. Expected one of {facilities}.")
    return facility


def get_health_facility_model_cls() -> type[HealthFacility]:
    return django_apps.get_model(get_health_facility_model())


def get_health_facility_model() -> str:
    return getattr(
        settings, "EDC_FACILITY_HEALTH_FACILITY_MODEL", "edc_facility.HealthFacility"
    )



# Generated by Django 6.0 on 2026-01-17 18:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0097_auto_20260115_0020"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalstock",
            name="returned",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="returned",
        ),
        migrations.AddField(
            model_name="historicalstock",
            name="destroyed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalstock",
            name="dispensed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalstock",
            name="in_transit",
            field=models.BooleanField(
                default=False,
                help_text="See stocktransferitem and confirmatlocationitem.",
            ),
        ),
        migrations.AddField(
            model_name="stock",
            name="destroyed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="stock",
            name="dispensed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="stock",
            name="in_transit",
            field=models.BooleanField(
                default=False,
                help_text="See stocktransferitem and confirmatlocationitem.",
            ),
        ),
    ]



# Generated by Django 5.2.6 on 2025-09-18 00:24

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_facility", "0016_alter_healthfacility_device_created_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="healthfacility",
            name="gps",
            field=models.CharField(
                blank=True,
                default="",
                help_text="copy and paste directly from google maps",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="healthfacility",
            name="report_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="healthfacility",
            name="title",
            field=models.CharField(blank=True, default="", max_length=150),
        ),
        migrations.AlterField(
            model_name="historicalhealthfacility",
            name="gps",
            field=models.CharField(
                blank=True,
                default="",
                help_text="copy and paste directly from google maps",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalhealthfacility",
            name="report_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalhealthfacility",
            name="title",
            field=models.CharField(blank=True, default="", max_length=150),
        ),
    ]



from edc_action_item.models import ActionModelMixin
from edc_consent.model_mixins import RequiresConsentFieldsModelMixin
from edc_model.models import BaseUuidModel
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin

from .action_items import SUBJECT_LOCATOR_ACTION
from .model_mixins import LocatorManager, LocatorModelMixin


class SubjectLocator(
    LocatorModelMixin,
    RequiresConsentFieldsModelMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    """A model completed by the user that captures participant
    locator information and permission to contact.
    """

    action_name = SUBJECT_LOCATOR_ACTION

    objects = LocatorManager()

    on_site = CurrentSiteManager()

    def natural_key(self):
        return (self.subject_identifier,)

    natural_key.dependencies = ("sites.Site",)

    class Meta(ActionModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Subject Locator"
        verbose_name_plural = "Subject Locators"
        indexes = (*ActionModelMixin.Meta.indexes, *BaseUuidModel.Meta.indexes)



from django import forms

from ...models import StorageBin, StorageBinItem


class StorageBinForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("capacity") is not None:
            item_count = StorageBinItem.objects.filter(
                storage_bin__bin_identifier=cleaned_data.get("bin_identifier")
            ).count()
            if item_count > 0 and cleaned_data.get("capacity") < item_count:
                raise forms.ValidationError(
                    {
                        "capacity": (
                            "Cannot be less than the number of items in bin. "
                            f"(Got {item_count})"
                        )
                    }
                )
        return cleaned_data

    class Meta:
        model = StorageBin
        fields = "__all__"



from .storage_bin_item import StorageBinItem


class StorageBinItemProxy(StorageBinItem):
    class Meta:
        proxy = True
        verbose_name = "Storage bin item"
        verbose_name_plural = "Storage bin items"



from __future__ import annotations

import contextlib
from datetime import datetime
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from django.apps import apps as django_apps
from django.db.models.deletion import ProtectedError

from edc_facility.exceptions import FacilityError
from edc_facility.utils import get_facility

from ..exceptions import CreateAppointmentError
from .appointment_creator import AppointmentCreator

if TYPE_CHECKING:
    from django.db.models import QuerySet

    from edc_consent.consent_definition import ConsentDefinition
    from edc_visit_schedule.schedule import Schedule
    from edc_visit_schedule.visit_schedule import VisitSchedule

    from ..models import Appointment


class AppointmentsCreator:
    """Note: Appointments are created using this class by
    the visit schedule.

    See also: edc_visit_schedule SubjectSchedule

    """

    appointment_creator_cls = AppointmentCreator

    def __init__(
        self,
        subject_identifier: str | None = None,
        visit_schedule: VisitSchedule | None = None,
        schedule: Schedule | None = None,
        report_datetime: datetime | None = None,
        appointment_model: str | None = None,
        site_id: int | None = None,
        skip_baseline: bool | None = None,
    ):
        self.subject_identifier: str = subject_identifier
        self.visit_schedule: VisitSchedule = visit_schedule
        self.schedule: Schedule = schedule
        self.report_datetime: datetime = report_datetime
        self.appointment_model: str = appointment_model
        self.site_id = site_id
        self.skip_baseline: bool | None = skip_baseline

    @property
    def appointment_model_cls(self) -> Appointment:
        return django_apps.get_model(self.appointment_model)

    def create_appointments(
        self,
        base_appt_datetime=None,
        taken_datetimes=None,
        skip_get_current_site: bool | None = None,
        consent_definition: ConsentDefinition | None = None,
    ) -> QuerySet[Appointment]:
        """Creates appointments when called by post_save signal.

        Timepoint datetimes are adjusted according to the available
        days in the facility.
        """
        taken_datetimes = taken_datetimes or []
        base_appt_datetime = (base_appt_datetime or self.report_datetime).astimezone(
            ZoneInfo("UTC")
        )

        timepoint_dates = self.schedule.visits_for_subject(
            subject_identifier=self.subject_identifier,
            report_datetime=base_appt_datetime,
            site_id=self.site_id,
            consent_definition=consent_definition,
        ).timepoint_dates(dt=base_appt_datetime)

        for visit, timepoint_datetime in timepoint_dates.items():
            try:
                facility = get_facility(visit.facility_name)
            except FacilityError as e:
                raise CreateAppointmentError(
                    f"{e} See {visit!r}. Got facility_name={visit.facility_name}"
                ) from e
            appointment = self.update_or_create_appointment(
                visit=visit,
                taken_datetimes=taken_datetimes,
                timepoint_datetime=timepoint_datetime,
                facility=facility,
                skip_get_current_site=skip_get_current_site,
            )
            taken_datetimes.append(appointment.appt_datetime)

        # check for existing appointment model instances after last timepoint
        try:
            last_timepoint = list(timepoint_dates.keys())[-1].timepoint
        except IndexError:
            pass
        else:
            self.delete_appointments_after_timepoint(last_timepoint)

        return self.appointment_model_cls.objects.filter(
            subject_identifier=self.subject_identifier,
            site_id=self.site_id,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
        ).order_by("timepoint")

    def update_or_create_appointment(self, **kwargs) -> Appointment:
        """Updates or creates an appointment for this subject
        for the visit.
        """
        opts = dict(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
            appointment_model=self.appointment_model,
            skip_baseline=self.skip_baseline,
            **kwargs,
        )
        appointment_creator = self.appointment_creator_cls(**opts)
        return appointment_creator.appointment

    def delete_unused_appointments(self) -> None:
        appointments = self.appointment_model.objects.filter(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
        )
        for appointment in appointments:
            with contextlib.suppress(ProtectedError):
                appointment.delete()

    def delete_appointments_after_timepoint(self, last_timepoint) -> None:
        """Delete appointments after a given timepoint.

        This is only relavent if the consent definition is extended
        by a consent definition extension.
        """
        for appointment in self.appointment_model_cls.objects.filter(
            subject_identifier=self.subject_identifier,
            site_id=self.site_id,
            timepoint__gt=last_timepoint,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
        ):
            with contextlib.suppress(ProtectedError):
                appointment.delete()



from django.conf import settings


def get_imp_schedule_names() -> tuple[list[str], list[str]]:
    imp_schedule_names = getattr(
        settings, "EDC_PHARMACY_IMP_VISIT_SCHEDULES", ["visit_schedule.schedule"]
    )
    visit_schedule_names = []
    schedule_names = []
    for imp_schedule_name in imp_schedule_names:
        v, s = imp_schedule_name.split(".")
        visit_schedule_names.append(v)
        schedule_names.append(s)
    return visit_schedule_names, schedule_names



from clinicedc_constants import NEW
from django.core.validators import MinValueValidator
from django.db import models
from edc_model.models import BaseUuidModel, HistoricalRecords
from sequences import get_next_value

from ...choices import ORDER_CHOICES
from .supplier import Supplier


class Manager(models.Manager):
    use_in_migrations = True


class Order(BaseUuidModel):
    order_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    order_datetime = models.DateTimeField(verbose_name="Order date/time")

    item_count = models.IntegerField(
        verbose_name="Item count",
        null=True,
        validators=[MinValueValidator(1)],
    )

    title = models.CharField(
        max_length=50,
        default="",
        blank=False,
        help_text="A short description of this order",
    )

    supplier = models.ForeignKey(
        Supplier,
        on_delete=models.PROTECT,
        verbose_name="Supplier",
        null=True,
        blank=False,
    )
    comment = models.TextField(default="", blank=True)

    sent = models.BooleanField(default=False)

    status = models.CharField(
        max_length=25,
        choices=ORDER_CHOICES,
        default=NEW,
        help_text="Updates in the signal",
    )

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.order_identifier}"

    def save(self, *args, **kwargs):
        if not self.order_identifier:
            self.order_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Order"
        verbose_name_plural = "Orders"



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms

from ..exceptions import RelatedVisitFieldError

if TYPE_CHECKING:
    from edc_crf.crf_form_validator import CrfFormValidator
    from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
    from edc_crf.modelform_mixins import InlineCrfModelFormMixin
    from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
    from edc_model.models import BaseUuidModel
    from edc_sites.model_mixins import SiteModelMixin
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    from ..modelform_mixins import VisitTrackingCrfModelFormMixin

    class RelatedVisitModel(SiteModelMixin, CreatesMetadataModelMixin, Base, BaseUuidModel):
        pass


__all__ = ["get_related_visit"]


def get_related_visit(
    modelform: (
        VisitTrackingCrfModelFormMixin
        | InlineCrfModelFormMixin
        | CrfFormValidator
        | CrfFormValidatorMixin
    ),
    related_visit_model_attr: str | None = None,
) -> RelatedVisitModel | None:
    """Returns the related visit model instance or None.

    Tries instance and cleaned data.
    """
    if related_visit_model_attr not in modelform.cleaned_data:
        try:
            related_visit = modelform.instance.related_visit
        except RelatedVisitFieldError:
            related_visit = None
        if not related_visit:
            raise forms.ValidationError(
                f"This field is required. Got `{related_visit_model_attr}.` (2)."
            )
    else:
        related_visit = modelform.cleaned_data.get(related_visit_model_attr)
    return related_visit



from django.core.management import color_style

from edc_utils.context_processors_check import edc_context_processors_check

style = color_style()


def context_processors_check(app_configs, **kwargs):
    return edc_context_processors_check(
        app_configs,
        app_label="edc_appointment",
        context_processor_name="edc_appointment.context_processors.constants",
        error_code=None,
        **kwargs,
    )



import inflect
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpRequest, HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator

from ..models import Stock, StorageBin, StorageBinItem
from .add_to_storage_bin_view import AddToStorageBinView, StorageBinError

p = inflect.engine()
MAX_STORAGE_BIN_CAPACITY = 50


def move_to_bin(
    storage_bin: StorageBin,
    stock_codes: list[str],
    user_modified: str,
    request: HttpRequest,
) -> tuple[list[str], list[str]]:
    codes_moved = []
    codes_not_moved = []
    # update storage bin capacity
    new_capacity = StorageBinItem.objects.filter(storage_bin=storage_bin).count() + len(
        stock_codes
    )
    if new_capacity > MAX_STORAGE_BIN_CAPACITY:
        raise StorageBinError(
            f"Storage bin {storage_bin.name} capacity "
            f"may not exceeded {MAX_STORAGE_BIN_CAPACITY}."
        )
    if new_capacity > storage_bin.capacity:
        storage_bin.capacity = new_capacity
        storage_bin.save()
        messages.add_message(
            request,
            messages.INFO,
            f"Storage bin {storage_bin.name} capacity has been increased to {new_capacity}.",
        )
    # try to move codes to target bin
    for code in stock_codes:
        try:
            stock_obj = Stock.objects.get(
                code=code,
                allocation__isnull=False,
                confirmed_at_location=True,
                stored_at_location=True,
                dispensed=False,
                location=storage_bin.location,
            )
        except ObjectDoesNotExist:
            stock_obj = None
            codes_not_moved.append(code)
        if stock_obj:
            try:
                obj = StorageBinItem.objects.get(stock=stock_obj)
            except ObjectDoesNotExist:
                codes_not_moved.append(code)
            else:
                obj.storage_bin = storage_bin
                obj.user_modified = user_modified
                obj.modified = timezone.now()
                obj.save()
                codes_moved.append(code)
    return codes_moved, codes_not_moved


@method_decorator(login_required, name="dispatch")
class MoveToStorageBinView(AddToStorageBinView):
    action_word = "Move"

    def redirect_on_stock_not_already_in_a_bin(
        self, stock_codes: list[str], storage_bin: StorageBin
    ) -> HttpResponseRedirect | None:
        if (
            stock_codes
            and Stock.objects.filter(
                code__in=stock_codes, storagebinitem__isnull=True
            ).exists()
        ):
            Stock.objects.filter(code__in=stock_codes, storagebinitem__isnull=True)
            messages.add_message(
                self.request,
                messages.ERROR,
                "Stock not found in any bin. Trying adding to storage bin first.",
            )
            url = reverse(
                "edc_pharmacy:move_to_storage_bin_url",
                kwargs={"storage_bin": storage_bin.id},
            )
            return HttpResponseRedirect(url)
        return None

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        stock_codes = request.POST.getlist("codes") if request.POST.get("codes") else None
        storage_bin = StorageBin.objects.get(id=kwargs.get("storage_bin"))
        items_to_scan = request.POST.get("items_to_scan") or kwargs.get("items_to_scan")
        if items_to_scan:
            items_to_scan = int(items_to_scan)

        self.redirect_on_has_duplicates(stock_codes, storage_bin)
        self.redirect_on_stock_not_already_in_a_bin(stock_codes, storage_bin)
        if items_to_scan and not stock_codes:
            url = reverse(
                "edc_pharmacy:move_to_storage_bin_url",
                kwargs={
                    "storage_bin": storage_bin.id,
                    "items_to_scan": items_to_scan,
                },
            )
            return HttpResponseRedirect(url)
        if items_to_scan and stock_codes:
            try:
                codes_moved, codes_not_moved = move_to_bin(
                    storage_bin, stock_codes, request.user.username, request
                )
            except StorageBinError as e:
                messages.add_message(request, messages.ERROR, str(e))
            else:
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    (
                        f"Moved {p.no('stock item', len(codes_moved))} "
                        f"to bin {storage_bin.name}. "
                        f"Skipped {len(codes_not_moved)}."
                    ),
                )
            return HttpResponseRedirect(self.storage_bin_changelist_url)
        url = reverse(
            "edc_pharmacy:move_to_storage_bin_url",
            kwargs={
                "storage_bin": storage_bin.id,
            },
        )
        return HttpResponseRedirect(url)



# Generated by Django 5.2.10 on 2026-01-25 02:21

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0123_remove_historicalrepackrequest_requested_qty_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalrepackrequest",
            name="repack_item_qty",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="Number of containers",
            ),
        ),
        migrations.AlterField(
            model_name="repackrequest",
            name="repack_item_qty",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="Number of containers",
            ),
        ),
    ]



from faker import Faker
from model_bakery.recipe import Recipe

from .models import CauseOfDeath

fake = Faker()

causeofdeath = Recipe(
    CauseOfDeath, display_name="cryptococcal_meningitis", name="cryptococcal_meningitis"
)



from django.contrib.admin import SimpleListFilter

__all__ = ["ScheduleStatusListFilter"]

from django.db.models import Count

from ..models import SubjectScheduleHistory


class ScheduleStatusListFilter(SimpleListFilter):
    title = "Schedule status"
    parameter_name = "schedule_status"

    def lookups(self, request, model_admin):  # noqa: ARG002
        names = []
        qs = (
            SubjectScheduleHistory.objects.values(
                "schedule_name", "onschedule_model", "offschedule_model"
            )
            .order_by("schedule_name", "onschedule_model", "offschedule_model")
            .annotate(cnt=Count("schedule_name"))
        )
        for s in ["on", "off"]:
            names.extend(
                [
                    (
                        f"{obj.get('schedule_name')}__{s}",
                        f"{s.title()}: {obj.get('schedule_name')}",
                    )
                    for obj in qs
                ]
            )
        return tuple(names)

    @property
    def subject_identifiers(self):
        schedule_name, status = self.value().split("__")
        return SubjectScheduleHistory.objects.filter(
            schedule_name=schedule_name, offschedule_datetime__isnull=status == "on"
        ).values_list("subject_identifier", flat=True)

    def queryset(self, request, queryset):  # noqa: ARG002
        if self.value() and self.value() != "none":
            queryset = queryset.filter(subject_identifier__in=self.subject_identifiers)
        return queryset



from django.conf import settings
from django_audit_fields import AUDIT_MODEL_FIELDS

DEFAULT_BASE_FIELDS = list(set(AUDIT_MODEL_FIELDS + ["id"]))
REPORT_DATETIME_FIELD_NAME = getattr(
    settings, "EDC_MODEL_REPORT_DATETIME_FIELD_NAME", "report_datetime"
)
